{****************************************************************************
*                                                                           *
*                                  ERRHNDLE                                 *
*                                                                           *
*                                                                           *
* Language:             FPC Pascal v0.99.10                                 *
* Environment:          IBM PC (MS-DOS)                                     *
* Supported extenders:  GO32V2                                              *
*                                                                           *
* Required switches:    -S2                                                 *
*                                                                           *
* Author:               Thomas Schatzl                                      *
* License:              LGPL, see the file COPYING.TXT, included in this    *
*                       distribution for details about the copyright.       *
* Date:                 06.02.1999                                          *
* Version:              rel0beta1                                           *
*                                                                           *
*      Send bug reports and feedback to tom_at_work@geocities.com           *
*                                                                           *
* Description:  Advanced generic error handling suited for error handling   *
*                                                                           *
*               NOTE: This error handler unit is still beta. I normally     *
*               don't release betas (not so early ones), but this time I    *
*               want to hear your comments about it.                        *
*                                                                           *
* See the history at the end too                                            *
*                                                                           *
*****************************************************************************}
unit errhndle;

{$MODE OBJFPC}

interface

const
     err_flag_terminate = $80000000;
{ error levels (to be implemented) }
(*     err_note, err_hint, err_warning, err_error, err_fatal *)

type
     error_id     = Longint;
     error_number = DWord;

     { an error message }
     perror_msg = ^error_msg;
     error_msg = record
               flags       : DWord;        { flags, unused }
               number      : error_number; { the error number to be written }
               description : ansistring;        { description of the error occured }
     end;

     { syntax definition of an error callback procedure which is called
       at an error. }
     error_callback = function (n : error_number) : error_msg;


     { standard 'unknown error' message template }
const
      error_unknown_msg : error_msg =
                       (flags  : 0;
                        number : 0;
                        description : 'Unknown error');

     { this variable holds the contents of the last occured error }
var
    errhndle_error : DWord;


{ main handler called at *every* error. Distributes error messages among
  the different registered modules }
procedure error_handler(i : error_id; e : error_number);

{ creates a new handler id for every module }
function createhandler(id_str : string; var id : error_id) : Boolean;

{ checks if the handler called s is registered and returns its id number if it
  is }
function queryhandlerid(s : string; var id : error_id) : Boolean;

{ sets a error procedure which is called when an error occurs. If this the
 given callback is nil, the handler is ignored }
function sethandler(id : error_id; p : error_callback) : Boolean;
{ retrieves error procedure of a specific module }
function gethandler(id : error_id; var p : error_callback) : Boolean;

implementation

{ error codes for this unit }
const
      err_handle_already_registered = $00010000;
      err_handler_id_not_found      = $00010001;

      { error messages }
      err_msglist : array[0..1] of error_msg = (
                  (flags : 0; number : err_handle_already_registered; description : 'Error Id already in use'),
                  (flags : 0; number : err_handler_id_not_found; description : 'Invalid error handler id'));

      { this units error id  }
var
    errhndle_error_id : error_id;

function errhndle_handler(n : error_number) : error_msg;
var i : DWord;
begin
     errhndle_error := n;
     i := 0;
     while ((err_msglist[i].number <> n) and (i <= high(err_msglist))) do i := i + 1;
     if (i = (high(err_msglist)+1)) then result := error_unknown_msg
     else result := err_msglist[i];
end;

{ the error handler implementation itself }

type { this is a single handlers data structure }
     phandler = ^handler;
     handler = record
             next, prev : phandler; { handlers are interconnected via a doubly linked list }
             id_string : string;    { the handlers' id string }
             id_number : error_id;  { the handlers' id number }
             callback  : error_callback; { the callback procedure called at an error }
             { to be added: 'error level' : level at which the callback operates
                            'exitcallback : callback called at an emergency exit
                                          (e.g. closes files, back to textmode, etc...
             }
     end;

{ pointer to the first handler data structure }
var
    pListfirst : phandler;
{ current value distributed as handler id's to modules }
    handlercount : DWord;

{ allocates memory for an empty handler (e.g. does nothing) and returns an
  identifier number to the caller.
}
function createhandler(id_str : string; var id : error_id) : Boolean;
var temp : phandler;
    newhandler : phandler;
    temp_id : error_id;
begin
     if not (queryhandlerid(id_str, temp_id)) then begin
        new(newhandler);
        {!!!!!! check for newhandler = nil }
        with (newhandler^) do begin
             next := nil;
             prev := nil;
             id_string := id_str;
             id_number := handlercount;
             callback := nil;
        end;

        id := handlercount;
        { add the new allocated handler at the end of the linked list }
        temp := plistfirst;
        if (temp <> nil) then begin
           while (temp^.next <> nil) do temp := temp^.next;

           newhandler^.prev := temp;
           temp^.next := newhandler;
        end else begin
        
            plistfirst := newhandler;
        end;
        { increase internal handler count }
        handlercount := handlercount + 1;
        result := true;
     end else begin
         result := false;
         {!!!!!!! ERROR : HANDLER ALREADY REGISTERED }
     end;
end;

{ retrieves the callback of the handler identified by the given error_id }
function gethandler(id : error_id; var p : error_callback) : Boolean;
var temp : phandler;
begin
     temp := plistfirst;
     if (temp<>nil) then begin
        while ((temp^.id_number <> id) and (temp^.next <> nil)) do
              temp := temp^.next;

        if (temp^.id_number = id) then begin
           p := temp^.callback;
           result := true;
        end else begin
            {!!!!  ERROR HANDLER ID NOT FOUND }
            result := false;
        end;
     end else begin
         { !!!!!! NO HANDLER REGISTERED AT ALL }
         result := false;
     end;
end;

{ sets the callback handler for the handler specified by id }
function sethandler(id : error_id; p : error_callback) : Boolean;
var temp : phandler;
begin
     temp := plistfirst;
     if (temp <> nil) then
        while ((temp^.id_number <> id) and (temp^.next <> nil)) do
              temp := temp^.next;
        
     if (temp^.id_number = id) then begin
        temp^.callback := p;
        result := true;
     end else begin
         result := false;
         {!!!!!! ERROR HANDLER ID NOT FOUND }
     end;
end;

function queryhandlerid(s : string; var id : error_id) : Boolean;
var temp : phandler;
begin
     temp := plistfirst;
     if (temp <> nil) then
        while ((temp^.next <> nil) and (temp^.id_string <> s)) do
              temp := temp^.next;
           
     if (temp^.id_string = s) then begin
        id := temp^.id_number;
        result := true;
     end else begin
         {!!!!!! ERROR HANDLER ID NOT FOUND }
         result := false;
     end;
end;

procedure error_handler(i : error_id; e : error_number);
var temp : phandler;

    msg : error_msg;
begin
     temp := plistfirst;
     if (temp <> nil) then
        while ((temp^.next <> nil) and (temp^.id_number <> i)) do
              temp := temp^.next;
     
     if (temp^.id_number = i) then begin
        if (pointer(temp^.callback) <> nil) then begin
           msg := temp^.callback(e);
           Writeln(stderr, 'Module ', temp^.id_string, ' reported an error ', hexstr(msg.number,8));
           Writeln(stderr, 'Description : ',msg.description);
           if ((msg.flags and err_flag_terminate)=err_flag_terminate) then begin
              { !!!!! DISPOSE LINKED LIST HERE !}
              halt(e);
           end;
        end;
     end else begin
         {!!!!!! ERROR HANDLER ID NOT FOUND }
     end;
end;

begin
     handlercount := 0;
     plistfirst := nil;
     createhandler('ERRHNDLE', errhndle_error_id);
     sethandler(errhndle_error_id, @errhndle_handler);
end.
{ 06-02-1999 start of history, first version
  - start of unit, first beta
}
