{#Z+}
{  Tools Library
   Copyright (c) 1993, 1997
   Cache Unit
   Programmers: Brad Williams, D.V.M.
   Version: 1.0.5 }
{#Z-}

unit Cache;

{ The tlCache provides an object-oriented system for implementing caches of
different types.  The objects are designed to be flexible with regards to
descendant implementation and symplicity of use. }

interface

uses

  Objects;

const

  coNilPointer = -3;
    { coNilPointer is used when attempting to insert a nil pointer into a
      cache when compiled using the #cdDebug# conditional define. }
    {#X TCache.Insert }


type

  {#Z+}
  PCachePage = ^TCachePage;
  {#Z-}
  TCachePage = Object(TObject)
    { TCachePage is a caching control object. It contains the necessary
      information to implement one page of a cache. The cache page is aged
      according to use, so more heavily used pages can be retained longer. }
    {#X TCache }
    Cycles: LONGINT;
      { Indicator used to represent many times the page has been accessed.

        Cycles can be used to determine which cache items to remove when new
        cache items must be loaded from the storage device. }
      {#X Age }
    Modified: Boolean;
      { True if the #Buffer# has been modified since read from the associated
        storage device. }
    Buffer: PChar;
      { Buffer is a pointer to the data buffered from the stream in this
        page. }
    Number: LongInt;
      { Number of the page.

        Number is normally relative to the start of the storage device so
        that two pages do not overlap in memory. }
    constructor Init (APageNumber: LongInt);
      { Creates an instance of a page with APageNumber.  Buffer is initially
        set to nil and should be set by the calling procedure.  If an error
        occurs Init fails.

        To minimize the amount of memory used by a cache item, the owning
        cache is responsible for allocating and deallocating Buffer based on
        its page size. }
    procedure Age;
      { Increases the age of the page by one cycle. }
      {#X Cycles UnAge }
    procedure UnAge;
      { Decreases the age of the page by one cycle. }
      {#X Cycles Age }
  end;  { of TCachePage }


  {#Z+}
  PCache = ^TCache;
  {#Z-}
  TCache = Object(TCollection)
    { A TCache holds pointers to #TCachePage# objects which buffer the data
      structures data. If an attempt is made to #Insert# a page and the cache
      is full, the #OldestPage# is deleted.

      The aging system for cache pages ensures the oldest, least accessed
      page is the first one released.  When any read or write to a cache
      page occurs, that page is UnAged.  When new pages are created they are
      #Insert#ed into the cache, which places them at the end of the cache.
      The first page encountered beginning at position 0 in the cache with
      the oldest age is the one deleted.  This way more used pages and newer
      pages "float" to the beginning of the cache, while less used older
      pages "sink" to the end of the cache.

      To ensure that a heavily used page at one point in time does not stay
      around too long after it is needed, all existing pages are aged when a
      new page is inserted.  The low page having the same age works because
      the most recently retrieved pages are Insert'ed at the end of the cache
      while the search for the youngest returns the youngest at the front of
      the cache.  Each access to a page should also cause the remainder of
      the pages to UnAge. }
    {#X TCachePage }
    constructor Init (APageSize: LONGINT; AMaxPages: LONGINT);
      { Initializes to cache to hold AMaxPages of APageSize bytes. }
    procedure FreeItem (Item: Pointer); virtual;
      { Disposes of the page Item and its associated buffer.

        Note: ClearCache does NOT #Flush# data to the stream. }
      {#X TCachePage }
    function Full: Boolean; virtual;
      { Returns True if Count is equal to #MaxPages#. }
    function GetPage (ANumber: LongInt): PCachePage; virtual;
      { Returns a pointer to the first page with Number = ANumber. }
      {#X PageInCache TCachePage }
    procedure Insert (Item: Pointer); virtual;
      { Inserts a new page into the cache aging all existing pages.

        Item must be a valid PCachePage.  When compiled using the #cdDebug#
        conditional define Insert checks for a nil Item and if so calls
        Error with Code set to coNilPointer and exits.

        If the cache is full when compiled using the #cdDebug# conditional
        define and Insert is called, Insert calls #Error# with Code set to
        coOverflow.  This situation should be prevented by the programmer
        and is considered a programmer error.  Removing the Debug conditional
        define for shipping versions is recommended to reduce code size and
        improve cache performance.

        Override this method if a stream is associated with the cache to
        flush the oldest page to the stream and Free it when the cache is
        full.  The new page, Item, can then be inserted. }
      {#X TStreamCache.AddPage }
    function NewPage (ANumber: LongInt): PCachePage; virtual;
      { Creates a new PCachePage using ANumber with an empty buffer of
        #PageSize# and returns a pointer to it. }
    function OldestPage: PCachePage; virtual;
      { Returns a pointer to the oldest, least used page in the cache. This
        should be the first page deleted when the cache is full. }
      {#X TCachePage.Age YoungestPage }
    function PageInCache (ANumber: LongInt): Boolean; virtual;
      { Returns True if ANumber is in a page in the cache based on
        #TCachePage.Number#. }
      {#X GetPage TCachePage }
   function ReadPage (ANumber: LongInt; var S: TStream): Boolean; virtual;
      { Reads #PageSize# bytes from the current position in stream S to
        Buffer of the page specified by ANumber and returns True.

        The page with ANumber should already be in the cache. When compiled
        with the #cdDebug# conditional define, ReadPage checks for the presence
        of a page with ANumber before attempting to read the data from the
        stream.  A call to ReadPage when the page specified by ANumber is not
        in the cache is considered a programmer error. It is recommended to
        compile the final shipping version without the Debug conditional
        define to decrease code size and improve cache performance.

        If an error occurs ReadPage returns False. }
      {#X ReadPage }
    function WritePage (ANumber: LongInt; var S: TStream): Boolean; virtual;
      { Writes #PageSize# bytes from Buffer of the page specified by ANumber
        to the current position in stream S and returns True.

        The page specified by ANumber is expected to be in the cache.  When
        compiled using the Debug conditional define, WritePage checks for a
        page existing in the cache with ANumber before attempting to write
        the data to the stream.  A call to WritePage when the page specified
        by ANumber is not in the cache is considered a programmer error. It
        is recommended to compile the final shipping version without the
        Debug conditional define to decrease code size and improve cache
        performance. }
      {#X ReadPage }
    function YoungestPage: PCachePage; virtual;
      { Returns a pointer to the youngest, most used page in the cache. }
      {#X TCachePage.Age OldestPage }
      private
    PageSize: LONGINT;
      { The size of the data in one cache page. }
      {#X TCachePage }
    MaxPages: longint;
      { The maximum number of pages which may be stored in the cache. }
      {#X PageSize TCachePage }
  end;  { of TCache }


  {#Z+}
  PStreamCache = ^TStreamCache;
  {#Z-}
  TStreamCache = Object(TStream)
    { TStreamCache implements a page caching mechanism for data written to
      or read from any #Stream#.  All #Read#s and #Write#s are buffered
      according to the frequency of their use.  Those pages more frequently
      accessed are maintained in memory for as long as possible.

      Streams with a recurring record structure such as database tables and
      indexes are ideal for cacheing to improve stream efficiency.  The use
      of #Offset# allows beginning caching at the first record in a table to
      make tables more efficient.  Data is written and read directly from
      the associated stream if an attempt is made to read/write to the cache
      at a position prior to offset.

      All data should be #Read#/#Write#, #Get#/#Put# to/from the cache rather
      than the stream itself.  Failing to do so may corrupt the stream and
      cache's data.

      In general, a TCache should not be used with a shared file since there
      is no gaurantee that the data on the stream will be in synchronization
      with the data in the cache. Flushing the cache before starting a child
      process is one exception. TStreamCache should also work well in a
      network server application. }
    {#X TCachePage }
    constructor Init (AStream: PStream; AOffset: LongInt;
                      APageSize: LONGINT; AMaxPages: LONGINT);
      { Intializes a cache on AStream beginning at Offset bytes from the
        start of the stream with AMaxPages of APageSize data in the cache. }
      {#X Done Stream TCache.PageSize TCache.MaxPages TCache TCachePage }
    destructor Done; virtual;
      { #Flush#es all modified cache pages to the stream and calls the
        inherited Done destructor.

        Calling Done without first making a successful call to #Flush# can
        result in loss of data if an error occurs during Flush. }
      {#X Init }
    procedure Flush; virtual;
      { Flushes all modified cache pages to the #Stream#. }
      {#X Done }
    function GetPos: LongInt; virtual;
      { Returns the current position of the internal pointer for the stream.

        GetPos is relative to the start of the stream, which is always 0. }
      {#X Seek CurPos }
    function GetSize: LongInt; virtual;
      { Returns the current size of the stream. An empty stream has a size
        of 0. }
      {#X Size GetPos }
    procedure Reset; virtual;
      { Resets the associated stream and updates Status and ErrorInfo to that
        of the associated stream. }
    procedure Seek (Pos: LongInt); virtual;
      { Moves the stream's internal pointer to Pos bytes from the start of
        the stream.  The start of a stream is position 0.

        The next #Get#, #Put#, #Read#, or #Write# will begin at Pos.

        If an error occurs Error is called with Code set to stError. }
      {#X GetPos CurPos }
    procedure Read (var Buf; Count: LONGINT); virtual;
      { Reads Count bytes of data from the current position in the stream
        and places it in Buf.

        If Count will cause a read beyond the end of a currently loaded cache
        page a new page is added to the cache with the appropriate
        information.  If the cache is full the oldest page is written to the
        stream and discarded to free memory for the new page. }
      {#X Write Cache TCachePage }
    procedure Truncate; virtual;
      { Flushes all modified cache pages to the associated #Stream# and
        disposes of them, then truncates the associated stream at the
        current position.

        The cache is empty after a call to Truncate. }
    procedure Write (var Buf; Count: LONGINT); virtual;
      { Writes Count bytes of data from Buf to the current position in the
        stream.

        If Count will cause a write beyond the end of a currently loaded
        cache page a new page is added to the cache with the appropriate
        information.  If the cache is full the oldest page is written to the
        stream and discarded to free memory for the new page. }
      {#X Write Cache TCachePage }
      private
    Stream: PStream;
      { Stream is the associated stream which is being cached.

        Once the stream has been passed to Init it should not be accessed
        directly until a call to #Done#.  Once Done is called the stream
        may be used as any other stream. This allows caching a stream while
        still being able to use it for other data if the stream will not be
        expanded once created, such as with resource and config files. }
      {#X Init Done }
    Offset: LongInt;
      { The offset from the beginning of the #Stream# where caching should
        begin.

        Since most structured data files have a header at the start, this
        allows not buffering the header information. }
    Cache: PCache;
      { The page cache itself.

        Cache contains pointers to TCachePages which buffer the data. }
      {#X TCachePage }
    Size: LongInt;
      { The length of the stream. This may be more than the actual length
        stored on #Stream# if the last page has not yet been written to
        Stream. }
      {#X GetSize }
    CurPos: LongInt;
      { The current position within the stream. }
      {#X GetPos Seek Read Write }
    function AddPage (APosition: LongInt): Boolean;
      { Calculates the page number based on #Offset# and Position, creates a
        new page with a blank buffer, and inserts it into the #Cache#. If the
        cache is full the oldest page is free'd prior to inserting the new
        page. }
      {#X FreePage LoadPage }
    function FreePage (APage: PCachePage): Boolean;
      { Flushes the buffer of APage to the #Stream#, calls TCache.Free to
        delete the page from the cache and free its memory, then returns
        True.

        If an error occurs APage is not deleted from the cache and FreePage
        returns False. }
      {#X AddPage }
    function LoadPage (APosition: LongInt): Boolean;
      { if the cache page specified by APosition is in the cache LoadPage
        returns True.

        If the page specified by APosition is not in the cache, LoadPage
        calls #AddPage# to attempt to create a new cache page and then loads
        it data from the stream. }
      {#X FreePage WritePage }
    function PageNumber (APosition: LongInt): LongInt;
      { Returns what would be #TCachePage.Number# for the specified position
        in the stream. The cache page does not have to be in memory when
        PageNumber is called. }
      {#X PageOffset TCache.PageInCache TCachePage }
    function PageOffset (Position: LongInt): LONGINT;
      { PageOffset returns what would be the offset of Position within the
        cache pages Buffer field for Position in the stream.  The page does
        not have to be in memory when PageOffset is called. }
      {#X PageNumber ReadFromCache WriteToCache }
    function ReadFromCache (var Buf; Count: LONGINT): Boolean;
      { ReadFromCache reads Count bytes from the current position in the
        cache to Buf and returns True.  The read operation may span several
        cache pages and require loading new pages from the stream. }
      {#X CurPos WriteToCache }
    function StreamOffset (Page: PCachePage): LongInt;
      { StreamOffset returns the position in the stream at which Page begins
        based on Page.Number. }
      {#X PageNumber }
    function WritePage (APage: PCachePage): Boolean;
      { Writes the page to the stream at StreamOffset(APage^.Number).

        APage must be a vlaid PCachePage. }
    function WriteToCache (var Buf; Count: longint): Boolean;
      { WriteToCache writes Count bytes from Buf to the current position in
        the in the cache and returns True.  The write operation may span
        several cache pages and require pages to be loaded from the stream or
        the stream expanded. }
      {#X CurPos ReadFromCache }
  end;  { of TCache }


implementation

const
  MaxByte      = 255;
  MaxShortInt  = 127;
  MaxINTEGER   = 32767;
  MaxLONGINT   = 2147483647;
{  MaxCardinal  = 2147483647*2+1;}
  MaxWORD         = 65535;
  MaxPointer   = 65520;
    { MaxPointer is the largest block that can be safely allocated on the
      heap, (64K-$8). }
  MaxPChar = Pred(Maxlongint);


const
  MinByte      =           0;
  MinShortInt  =        -128;
  MinInteger   =      -32768;
  MinLongint   = -2147483647;
  MinWORD      =           0;

type
  Str2   = string[2];
  Str4   = string[4];
  Str5   = string[5];
  Str8   = string[8];
  Str11  = string[11];
  Str25  = String[25];
  Str64  = string[64];

type
  PWORD    =^WORD;
  Plongint = ^longint;
  PByte    = ^Byte;
  PReal    = ^Real;

type
  IntRec = record
    Hi: Byte;
    Lo: Byte;
  end;

type
  TActionProcedure = procedure (Item : Pointer);
  TActionProc = procedure (Item : Pointer; ACallerFrame : longint);
  TTestFunction = function (Item : Pointer) : Boolean;
  TTestFunc = function (Item : Pointer; ACallerFrame : longint) : Boolean;


{****************************************************************************}
{                     Local Procedures and Functions                         }
{****************************************************************************}

{****************************************************************************}
{ TCache Object                                                              }
{****************************************************************************}
{****************************************************************************}
{ TCache.Init                                                                }
{****************************************************************************}
constructor TCache.Init (APageSize: longint; AMaxPages: longint);
begin
  if (not TCollection.Init(AMaxPages,1)) then
    Fail;
  PageSize := APageSize;
  MaxPages := AMaxPages;
end;

{****************************************************************************}
{ TCache.FreeItem                                                            }
{****************************************************************************}
procedure TCache.FreeItem (Item: Pointer);
begin
  if (Item <> nil) then
    FreeMem(PCachePage(Item)^.Buffer,PageSize);
  Dispose(PCachePage(Item),Done);
end;

{****************************************************************************}
{ TCache.Full                                                                }
{****************************************************************************}
function TCache.Full: Boolean;
begin
  Full := (Count = MaxPages);
end;

{****************************************************************************}
{ TCache.GetPage                                                             }
{****************************************************************************}
function TCache.GetPage (ANumber: LongInt): PCachePage;
  {******************************************************}
  function CorrectPage (Page: PCachePage): Boolean;
  begin
    CorrectPage := (Page^.Number = ANumber);
  end;
  {******************************************************}
begin
  GetPage := FirstThat(@CorrectPage);
end;

{****************************************************************************}
{ TCache.Insert                                                              }
{****************************************************************************}
procedure TCache.Insert (Item: Pointer);
{$ifdef cdDebug}
var
  Dummy: longint;
{$endif cdDebug}
  {*****************************************}
  procedure AgePage (Page: PCachePage);
  begin
    Page^.Age;
  end;
  {*****************************************}
begin
  {$ifdef cdDebug}
  if (Item = nil) then
  begin
    Dummy := 0;
    Error(coNilPointer,Dummy);
    Exit;
  end;
  if (Count = MaxPages) then
  begin
    Error(coOverflow,Count);
    Exit;
  end;
  {$endif cdDebug}
  ForEach(@AgePage);
  TCollection.Insert(Item);
end;

{****************************************************************************}
{ TCache.NewPage                                                             }
{****************************************************************************}
function TCache.NewPage (ANumber: LongInt): PCachePage;
var
  Page: PCachePage;
begin
  Page := New(PCachePage,Init(ANumber));
  if (Page <> nil) then
  begin
    GetMem(Page^.Buffer,PageSize);
    if (Page^.Buffer = nil) then
      Dispose(Page,Done);
  end;
  NewPage := Page;
end;

{****************************************************************************}
{ TCache.OldestPage                                                          }
{****************************************************************************}
function TCache.OldestPage: PCachePage;
var
  Page: PCachePage;
  {**********************************************************}
  function FindOldestPage (APage: PCachePage): Boolean;
  begin
    if APage^.Cycles > Page^.Cycles then
      Page := APage;
  end;
  {**********************************************************}
begin
  if (Count = 0) then
  begin
    OldestPage := nil;
    Exit;
  end;
  Page := PCachePage(At(0));
  ForEach(@FindOldestPage);
  OldestPage := Page;
end;

{****************************************************************************}
{ TCache.PageInCache                                                         }
{****************************************************************************}
function TCache.PageInCache (ANumber: LongInt): Boolean;
begin
  PageInCache := (GetPage(ANumber) <> nil);
end;

{****************************************************************************}
{ TCache.ReadPage                                                            }
{****************************************************************************}
function TCache.ReadPage (ANumber: LongInt; var S: TStream): Boolean;
var
  Page: PCachePage;
begin
  Page := GetPage(ANumber);
  {$ifdef cdDebug}
  if Page = nil
     then begin
            ReadPage := False;
            Exit;
          end;
  {$endif cdDebug}
  S.Read(Page^.Buffer^,PageSize);
  ReadPage := (S.Status = stOk);
end;

{****************************************************************************}
{ TCache.WritePage                                                           }
{****************************************************************************}
function TCache.WritePage (ANumber: LongInt; var S: TStream): Boolean;
var
  Page: PCachePage;
begin
  Page := GetPage(ANumber);
  {$ifdef cdDebug}
  if Page = nil
     then begin
            WritePage := False;
            Exit;
          end;
  {$endif cdDebug}
  S.Write(Page^.Buffer^,PageSize);
  WritePage := (S.Status = stOk);
end;

{****************************************************************************}
{ TCache.YoungestPage                                                        }
{****************************************************************************}
function TCache.YoungestPage: PCachePage;
var
  Page: PCachePage;
  {************************************************************}
  function FindYoungestPage (APage: PCachePage): Boolean;
  begin
    if APage^.Cycles < Page^.Cycles then
      Page := APage;
  end;
  {************************************************************}
begin
  if Count = 0 then
  begin
    YoungestPage := nil;
    Exit;
  end;
  Page := PCachePage(At(0));
  ForEach(@FindYoungestPage);
  YoungestPage := Page;
end;

{****************************************************************************}
{ TCachePage Object                                                          }
{****************************************************************************}
{****************************************************************************}
{ TCachePage.Init                                                            }
{****************************************************************************}
constructor TCachePage.Init (APageNumber: LongInt);
begin
  if (not TObject.Init) then
    Fail;
  Number := APageNumber;
  Buffer := nil;
  Modified := False;
  Cycles := Maxlongint div 2;
end;

{****************************************************************************}
{ TCachePage.Age                                                             }
{****************************************************************************}
procedure TCachePage.Age;
begin
  if Cycles < Maxlongint then
    Inc(Cycles);
end;

{****************************************************************************}
{ TCachePage.UnAge                                                           }
{****************************************************************************}
procedure TCachePage.UnAge;
begin
  if Cycles > 0 then
    Dec(Cycles);
end;

{****************************************************************************}
{ TStreamCache Object                                                        }
{****************************************************************************}

{****************************************************************************}
{ TStreamCache.Init                                                          }
{****************************************************************************}
constructor TStreamCache.Init (AStream: PStream; AOffset: LongInt;
                               APageSize: longint; AMaxPages: longint);
begin
  if (AStream = nil) or (AStream^.Status <> stOk) or (not TStream.Init) then
    Fail;
  Cache := New(PCache,Init(APageSize,AMaxPages));
  if Cache = nil then
    TStream.Done;
  Stream := AStream;
  Offset := AOffset;
  Status := stOk;
  ErrorInfo := 0;
end;

{****************************************************************************}
{ TStreamCache.Done                                                          }
{****************************************************************************}
destructor TStreamCache.Done;
begin
  Flush;
  Dispose(Cache,Done);
  TStream.Done;
end;

{****************************************************************************}
{ TStreamCache.AddPage                                                       }
{****************************************************************************}
function TStreamCache.AddPage (APosition: LongInt): Boolean;
var
  Page: PCachePage;
begin
  AddPage := False;
    { make sure there's room in the cache for a new page }
  if Cache^.Full and (not FreePage(Cache^.OldestPage)) then
    Exit;
    { try to make a new cache page }
  repeat
    Page := Cache^.NewPage(PageNumber(APosition));
    if (Page = nil) then  { insufficient memory }
      FreePage(Cache^.OldestPage);  { try to free the oldest page }
    { until success, error or no more pages to free }
  until (Page <> nil) or (Status <> stOk) or (Cache^.Count = 0);
  if (Page <> nil) then
  begin
    Cache^.Insert(Page);
    AddPage := True;
  end;
end;

{****************************************************************************}
{ TStreamCache.Flush                                                         }
{****************************************************************************}
procedure TStreamCache.Flush;
  {*****************************************************}
  procedure FlushItem (Page: PCachePage);
  begin
    if Page^.Modified then
    begin
      Stream^.Seek(StreamOffset(Page));
      Stream^.Write(Page^.Buffer^,Cache^.PageSize);
      Page^.Modified := (Stream^.Status <> stOk);
    end;
  end;
  {*****************************************************}
begin
  Cache^.ForEach(@FlushItem);
  Status := Stream^.Status;
  ErrorInfo := Stream^.ErrorInfo;
end;

{****************************************************************************}
{ TStreamCache.FreePage                                                      }
{****************************************************************************}
function TStreamCache.FreePage (APage: PCachePage): Boolean;
begin
  FreePage := False;
  {$ifdef cdDebug}
  if APage = nil
     then Exit;
  {$endif cdDebug}
  if (not WritePage(APage)) then
    Exit;
  Cache^.Free(APage);
end;

{****************************************************************************}
{ TStreamCache.GetPos                                                        }
{****************************************************************************}
function TStreamCache.GetPos: LongInt;
begin
  GetPos := CurPos;
end;

{****************************************************************************}
{ TStreamCache.GetSize                                                       }
{****************************************************************************}
function TStreamCache.GetSize: LongInt;
begin
  GetSize := Size;
end;

{****************************************************************************}
{ TStreamCache.LoadPage                                                      }
{****************************************************************************}
function TStreamCache.LoadPage (APosition: LongInt): Boolean;
begin
  LoadPage := True;
  if Cache^.PageInCache(PageNumber(APosition)) then
    Exit;
  if AddPage(APosition) then  { create a new cache page to hold the data }
  begin  { load the data from the stream into the new page }
    Stream^.Seek(APosition);
    Stream^.Read(PCachePage(Cache^.At(Pred(Cache^.Count)))^.Buffer[0],
                 Cache^.PageSize);
    Status := Stream^.Status;
    ErrorInfo := Stream^.ErrorInfo;
    if Status = stOk then
    begin  { only update CurPos if no errors }
      CurPos := APosition;
      LoadPage := True;
    end
    else Cache^.AtFree(Pred(Cache^.Count));
  end
  else LoadPage := False;
end;

{****************************************************************************}
{ TStreamCache.PageNumber                                                    }
{****************************************************************************}
function TStreamCache.PageNumber (APosition: LongInt): LongInt;
begin
  PageNumber := (APosition - Offset) div Cache^.PageSize;
end;

{****************************************************************************}
{ TStreamCache.PageOffset                                                    }
{****************************************************************************}
function TStreamCache.PageOffset (Position: LongInt): longint;
begin
  PageOffset := Position - (PageNumber(Position) * Cache^.PageSize);
end;

{****************************************************************************}
{ TStreamCache.Read                                                          }
{****************************************************************************}
procedure TStreamCache.Read (var Buf; Count: longint);
var
  BytesNotCached: LongInt;
  Data: array[0..Pred(Maxlongint)] of Char absolute Buf;
  {$ifdef cdDebug}
  Page: PCachePage;
  {$endif cdDebug}
begin
  if Status <> stOk then
    Exit;
  if Count > (Size - CurPos) then  { read attempt beyond end of stream }
  begin
    Error(stReadError,longint(Count - (Size - CurPos)));
    Exit;
  end;
  if CurPos < Offset then  { read non-cached data }
  begin
    BytesNotCached := (Offset - CurPos);
    Stream^.Seek(CurPos);  { keep stream in synch with cache }
    if BytesNotCached > Count then
      Stream^.Read(Buf,Count)
    else Stream^.Read(Buf,longint(BytesNotCached));
    if Stream^.Status = stOk then
      if BytesNotCached > Count then
        Inc(CurPos,Count)
      else Inc(CurPos,BytesNotCached);
  end
  else BytesNotCached := 0;
    { read remaining bytes from cache }
  if (Stream^.Status = stOk) and (Count > BytesNotCached) then
    ReadFromCache(Data[BytesNotCached],longint(Count - BytesNotCached));
  Status := Stream^.Status;
  ErrorInfo := Stream^.ErrorInfo;
end;

{****************************************************************************}
{ TStreamCache.ReadFromCache                                                 }
{****************************************************************************}
function TStreamCache.ReadFromCache (var Buf; Count: longint): Boolean;
var
  Data: array[0..Pred(Maxlongint)] of Char absolute Buf;
  BytesRead: longint;
  BytesToRead: longint;
begin
  ReadFromCache := False;
  BytesRead := 0;
  BytesToRead := (Cache^.PageSize - PageOffset(CurPos));
  while (Status = stOk) and LoadPage(CurPos) and (BytesRead < Count) do
  begin  { read the desired page's data }
    Move(PCachePage(Cache^.GetPage(CurPos))^.Buffer[PageOffset(CurPos)],
         Data[BytesRead],BytesToRead);
    Inc(BytesRead,BytesToRead);
    Inc(CurPos,BytesToRead);
    BytesToRead := Count - BytesRead;
    if BytesToRead > Cache^.PageSize then
      BytesToRead := Cache^.PageSize;
  end;
  ReadFromCache := (Status = stOk);
end;

{****************************************************************************}
{ TStreamCache.Reset                                                         }
{****************************************************************************}
procedure TStreamCache.Reset;
begin
  Stream^.Reset;
  Status := Stream^.Status;
  ErrorInfo := Stream^.ErrorInfo;
end;

{****************************************************************************}
{ TStreamCache.Seek                                                          }
{****************************************************************************}
procedure TStreamCache.Seek (Pos: LongInt);
begin
  if Status <> stOk then
    Exit;
  if (Pos > Size) then
  begin
    Error(stReadError,ErrorInfo);
    Exit;
  end;
  if Cache^.PageInCache(PageNumber(Pos)) then
    CurPos := Pos
  else if not LoadPage(Pos) then
         Error(stReadError,ErrorInfo);
end;

{****************************************************************************}
{ TStreamCache.StreamOffset                                                  }
{****************************************************************************}
function TStreamCache.StreamOffset (Page: PCachePage): LongInt;
begin
  StreamOffset := (Page^.Number * Cache^.PageSize) + Offset;
end;

{****************************************************************************}
{ TStreamCache.Truncate                                                      }
{****************************************************************************}
procedure TStreamCache.Truncate;
var
  CurrentPosition: LongInt;
begin
  CurrentPosition := CurPos;
  Flush;
  Seek(CurrentPosition);
  Stream^.Truncate;
  Status := Stream^.Status;
  ErrorInfo := Stream^.ErrorInfo;
end;

{****************************************************************************}
{ TStreamCache.Write                                                         }
{****************************************************************************}
procedure TStreamCache.Write (var Buf; Count: longint);
var
  BytesNotCached: LongInt;
  {$ifdef cdDebug}
  Page: PCachePage;
  {$endif cdDebug}
begin
  if Status <> stOk then
    Exit;
  if Count > (Size - CurPos) then
  begin
    Error(stWriteError,Count - (Size - CurPos));
    Exit;
  end;
  if CurPos < Offset then  { check for non-cached data }
    BytesNotCached := (Offset - CurPos)
  else BytesNotCached := 0;
  if BytesNotCached > 0 then  { write non-cached data }
  begin
    Stream^.Seek(CurPos);  { keep stream in synch with cache }
    if BytesNotCached > Count then
      Stream^.Write(Buf,Count)
    else Stream^.Write(Buf,longint(BytesNotCached));
    if Stream^.Status = stOk then
      Inc(CurPos,Maxlongint);
  end;
    { write remaining cached data }
  if (Stream^.Status = stOk) and (Count > BytesNotCached) then
    WriteToCache(Ptr(Seg(Buf),longint(Ofs(Buf) + BytesNotCached))^,
                 longint(Count - BytesNotCached));
  Status := Stream^.Status;
  ErrorInfo := Stream^.ErrorInfo;
end;

{****************************************************************************}
{ TStreamCache.WritePage                                                     }
{****************************************************************************}
function TStreamCache.WritePage (APage: PCachePage): Boolean;
  { APage must always be a valid PCachePage. }
var
  Pos: LongInt;
begin
  if APage^.Modified then
  begin
    Pos := StreamOffset(APage);
    Stream^.Seek(Pos);
      { check for write of unused data past end of stream }
    if (Pos + Cache^.PageSize) > Size then
    begin
      { check for more than one page not written to Stream }
      { recursively write other pages first? or just call Flush?
        or something else? }
      { write this page }
      Stream^.Write(APage^.Buffer^,(Size - Pos));
    end
    else Stream^.Write(APage^.Buffer^,Cache^.PageSize);
    Status := Stream^.Status;
    ErrorInfo := Stream^.ErrorInfo;
    APage^.Modified := (Status <> stOk);
  end;
  WritePage := APage^.Modified;
end;

{****************************************************************************}
{ TStreamCache.WriteToCache                                                  }
{****************************************************************************}
function TStreamCache.WriteToCache (var Buf; Count: longint): Boolean;
var
  Data: array[0..Pred(Maxlongint)] of Char absolute Buf;
  BytesWritten: longint;
  BytesToWrite: longint;
begin
  WriteToCache := False;
  BytesWritten := 0;
  BytesToWrite := (Cache^.PageSize - PageOffset(CurPos));
  while (Status = stOk) and LoadPage(CurPos) and (BytesWritten < Count) do
  begin  { Write the desired page's data }
    Move(Data[BytesWritten],
         PCachePage(Cache^.GetPage(CurPos))^.Buffer[PageOffset(CurPos)],
         BytesToWrite);
    Inc(BytesWritten,BytesToWrite);
    Inc(CurPos,BytesToWrite);
    BytesToWrite := Count - BytesWritten;
    if BytesToWrite > Cache^.PageSize then
      BytesToWrite := Cache^.PageSize;
  end;
  WriteToCache := (Status = stOk);
end;

{****************************************************************************}
{                    Global Procedures and Functions                         }
{****************************************************************************}

{****************************************************************************}
{               Initialization of Global and Local Variables                 }
{****************************************************************************}

begin
end.
