{****************************************************************************
*                                                                           *
*                            The Memory Unit                                *
*                                                                           *
*                                                                           *
* Language:             FPC Pascal v0.9.11+ (snapshot from 13.5.)           *
* Environment:          IBM PC (I386)                                       *
* Supported extenders:  GO32V1, GO32V2, PMODE/DJ                            *
*                                                                           *
* Required switches:    none                                                *
*                                                                           *
* Author:               Thomas Schatzl                                      *
* License:              LGPL, see the file COPYING.TXT, included in this    *
*                       distribution for details about the copyright.       *
* Date:                 13.05.1999                                          *
* Version:              rel2rev6                                            *
*                                                                           *
*      Send bug reports and feedback to tom_at_work@geocities.com           *
*                                                                           *
* Description:  New fast memory copy, memory set functions (up to four times*
*               as fast as go32 functions on Intel Pentium). Uses FPU       *
*               or MMX instructions if advantageous (FPU only on original   *
*               Intel Pentium non-MMX CPUs).                                *
*               Experimental support for KNI instructions added.            *
*               Boolean operations between two memory areas or memory and a *
*               single value possible. Fast autodetecting MMX support for   *
*               boolean operations. All 16 possible ROPs (raster operations)*
*               available, matching Micro$oft order.                        *
*                                                                           *
*               Automatically writes some information at startup to stderr  *
*               if a global compiler directive DEBUG is specified           *
*                                                                           *
*               Important : MUST be compiled with a GNU Assembler v2.9.1 or *
*               higher to work correctly. The standard assembler included   *
*               with FPC 0.99.10 is v2.8.1, which doesn't assemble one MMX  *
*               instruction properly. You may get the DOS version from my   *
*               homepage.                                                   *
*                                                                           *
* See also at the history at the end                                        *
*                                                                           *
*****************************************************************************}

unit Memory;

interface

{$OUTPUT_FORMAT AS}

{$DEFINE USE_KNI} { Undefine this to remove KNI support since its not yet
					tested }

type

{ This enumeration defines the possible bit operations between src and dst }

     Mem_Op = (MEM_CLEAR := 0,     { dst := 0 }
               MEM_NOT_DSTORSRC,   { dst := not (dst or src) }
               MEM_DST_AND_NOTSRC, { dst := dst and (not src) }
               MEM_NOTSRC,         { dst := not src }
               MEM_SRC_AND_NOTDST, { dst := src and (not dst) }
               MEM_NOTDST,         { dst := not dst }
               MEM_DST_XOR_SRC,    { dst := dst xor src }
               MEM_NOT_DSTANDSRC,  { dst := not (dst and src) }
               MEM_DST_AND_SRC,    { dst := dst and src }
               MEM_NOT_DSTXORSRC,  { dst := not (dst xor src) }
               MEM_DST,            { dst := dst; do nothing }
               MEM_DST_OR_NOTSRC,  { dst := dst or (not src) }
               MEM_SRC,            { dst := src; copy }
               MEM_NOTDST_OR_SRC,  { dst := (not dst) or src }
               MEM_DST_OR_SRC,     { dst := dst or src }
               MEM_SET);           { dst := 1 }

{ copies a block from src to dst }
procedure memcpy (var src, dst; size : DWord);

{ do boolean operation between src and dst and store the result in dst. If
  source and dest overlap the result is undefined
}
procedure memchange (var src, dst; size : DWord; op : Mem_Op);

{ Boolean operation between memory and a single value.
}
procedure memchangeValueB (var x; size : DWord; value : Byte; op : Mem_Op);
procedure memchangeValueW (var x; size : DWord; value : Word; op : Mem_Op);
procedure memchangeValueD (var x; size : DWord; value : DWord; op : Mem_Op);

{ set a memory area to a specific value, the same as fillchar. Enhanced
  because FPU is used, and so there's a speed gain of up to 100% comparing
  to the original procedure.
}
procedure memsetB (var x; size : DWord; value : Byte);
procedure memsetW (var x; size : DWord; value : Word);
procedure memsetD (var x; size : DWord; value : DWord);

{ same as seg_move, copies memory from different memory locations to
  another across selector boundaries. When the 2 memory areas overlap, the
  procedure copies the memory the right way; the unit automatically decides
  whether to use FPU or standard instructions depending on processor type }
var seg_memcpy : procedure (srcseg : Word; srcofs : DWord; dstseg : Word; dstofs : DWord; size : DWord);

{ this variable holds the fastest way to change memory contents. It is
  automatically decided at startup which routine it holds, either
  seg_memchangeSTD (normal opcodes) or seg_memchangeMMX (uses MMX
  extensions if available). The floating point unit is cleared after MMX
  use }
var seg_memchange : procedure (srcsel : Word; srcofs : DWord; dstsel : Word; dstofs : DWord; size : DWord; op : Mem_Op);

{ Boolean operation of a memory block, automatically uses MMX if available }
var seg_memchangeValueB : procedure (sel : Word; ofs : DWord; size : DWord; value : Byte; op : Mem_Op);
var seg_memchangeValueW : procedure (sel : Word; ofs : DWord; size : DWord; value : Word; op : Mem_Op);
var seg_memchangeValueD : procedure (sel : Word; ofs : DWord; size : DWord; value : DWord; op : Mem_Op);

{ sets a memory area to a specific value, either byte/word/dword sized. At
  startup it is automatically decided wheter it uses FPU or not for maximum
  speed enhancement }
var seg_memsetB : procedure (seg : Word; ofs : DWord; size : DWord; value : Byte);
var seg_memsetW : procedure (seg : Word; ofs : DWord; size : DWord; value : Word);
var seg_memsetD : procedure (seg : Word; ofs : DWord; size : DWord; value : DWord);

implementation

{$ASMMODE DIRECT}
{$PACKRECORDS 4}

uses cpu;

var mydseg : Word;

{ -------------------------------------------------------------------------- }
{ FPU versions of :                                                          }
{                  seg_memcpy() (last rev 14.02.1999)                        }
{ -------------------------------------------------------------------------- }

{ copies size bytes of memory between src and dst; FPU version }
procedure seg_memcpy_fpu(srcseg : Word; srcofs : DWord; dstseg : Word; dstofs : DWord; size : DWord); assembler;
asm
   .ALIGN 4
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .LFPU_SEGMCPY_EXIT        { exit if size = 0 }

   movw SRCSEG, %ax
   movw DSTSEG, %bx
   movl SRCOFS, %esi
   movl DSTOFS, %edi

   pushw %ds                    { prepare segment registers }
   pushw %es
   movw %ax, %ds
   movw %bx, %es

   cmpw %ax, %bx
   jne .LFPU_SEGMCPY_FORWARD    { if (srcseg <> dstseg) then copy forward }
   cmpl %esi, %edi
   je .LFPU_SEGMCPY_FINISH      { exit if srcseg:srcofs equals dstseg:dstofs }
   jl .LFPU_SEGMCPY_FORWARD     { or (dstofs < srcofs) then copy forward }

   movl %esi, %edx
   addl %ecx, %edx

   cmpl %edi, %edx              { or ((srcofs+size) < dstofs) then copy forward }
   jl .LFPU_SEGMCPY_FORWARD

.LFPU_SEGMCPY_BACKWARD:
   std                          { prepare for backward copy }

   movl %ecx, %edx              { save size }

   addl %ecx, %esi              { move offsets to end of block }
   addl %ecx, %edi

   movl %esi, %ebx              { get first number of bytes for alignment to }
                                { source }
   andl $7, %ebx

   cmpl %ebx, %ecx
   jle .LFPU_SEGMCPY_BW_LASTBYTES
   
   movl %ebx, %ecx
   decl %esi; decl %edi
   
   rep
   movsb
   
   incl %esi; incl %edi

   subl %ebx, %edx               { adjust size (subtract first alignment bytes) }
   movl %edx, %ecx

.LFPU_SEGMCPY_BW_QWORD:
   
   shrl $3, %ecx
   jz .LFPU_SEGMCPY_BW_LASTBYTES

   testl $1, %ecx
   jz .LFPU_SEGMCPY_BW_NOFIRST8
   subl $8, %esi
   subl $8, %edi
   fildq (%esi)
   fistpq %es:(%edi)

.LFPU_SEGMCPY_BW_NOFIRST8:
   shrl $1, %ecx
   
.LFPU_SEGMCPY_BW_QWORD_LOOP16:
   decl %ecx
   js .LFPU_SEGMCPY_BW_NOLOOP16
   subl $16, %esi
   subl $16, %edi
   fildq (%esi)
   fildq 8(%esi)
   fxch
   fistpq %es:(%edi)
   fistpq %es:8(%edi)
   jmp .LFPU_SEGMCPY_BW_QWORD_LOOP16

.LFPU_SEGMCPY_BW_NOLOOP16:
   movl %edx, %ecx
   
.LFPU_SEGMCPY_BW_LASTBYTES:
   andl $7, %ecx
   decl %esi; decl %edi

   rep
   movsb

   jmp .LFPU_SEGMCPY_FINISH

.LFPU_SEGMCPY_FORWARD:
   cld                          { prepare for forward copy }

   movl %ecx, %edx              { save size }

   movl $8, %ebx
   subl %esi, %ebx              { align on srcseg:srcofs }
   andl $7, %ebx
   jz .LFPU_SEGMCPY_FWD_QWORD   { ebx = number of bytes of first alignment }

   cmpl %ebx, %ecx              { if (size <= ebx) then goto lastbytes }
   jle .LFPU_SEGMCPY_FWD_LASTBYTES

   movl %ebx, %ecx              { prepare to move first alignment bytes }
   rep
   movsb
   
   subl %ebx, %edx
   movl %edx, %ecx              { decrease size by first alignment }
   
.LFPU_SEGMCPY_FWD_QWORD:

   shrl $3, %ecx
   jz .LFPU_SEGMCPY_FWD_LASTBYTES { if (size<8) then goto lastbytes  }

   testl $1, %ecx			    { copy a single 8 byte block if needed }
   jz .LFPU_SEGMCPY_FWD_NOFIRST8
   fildq (%esi)
   fistpq %es:(%edi)
   addl $8, %esi
   addl $8, %edi

.LFPU_SEGMCPY_FWD_NOFIRST8:
   shrl $1, %ecx                { shr count by 1 since we copy in 16 byte blocks }

.LFPU_SEGMCPY_FWD_QWORD_LOOP16:
   decl %ecx
   js .LFPU_SEGMCPY_FWD_NOLOOP16
   fildq (%esi)
   fildq 8(%esi)
   fxch
   fistpq %es:(%edi)
   fistpq %es:8(%edi)
   addl $16, %esi
   addl $16, %edi
   jmp .LFPU_SEGMCPY_FWD_QWORD_LOOP16

.LFPU_SEGMCPY_FWD_NOLOOP16:

   movl %edx, %ecx

.LFPU_SEGMCPY_FWD_LASTBYTES:
   andl $7, %ecx                { copy last alignment bytes }

   rep
   movsb

.LFPU_SEGMCPY_FINISH:
   popw %es                     { restore segment registers }
   popw %ds
   
.LFPU_SEGMCPY_EXIT:
end;

{ -------------------------------------------------------------------------- }
{ MMX versions of :                                                          }
{                  seg_memcpy()  (last rev 02.01.1999)                       }
{                  seg_memsetb() (last rev 02.01.1999)                       }
{                  seg_memsetw() (last rev 02.01.1999)                       }
{                  seg_memsetl() (last rev 02.01.1999)                       }
{ -------------------------------------------------------------------------- }

procedure seg_memcpy_mmx(srcseg : Word; srcofs : DWord; dstseg : Word; dstofs : DWord; size : DWord); assembler;
asm
   .ALIGN 4
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .LMMX_SEGMCPY_EXIT        { exit if size = 0 }

   movw SRCSEG, %ax
   movw DSTSEG, %bx
   movl SRCOFS, %esi
   movl DSTOFS, %edi

   pushw %ds                    { prepare segment registers }
   pushw %es
   movw %ax, %ds
   movw %bx, %es

   cmpw %ax, %bx
   jne .LMMX_SEGMCPY_FORWARD    { if (srcseg <> dstseg) then copy forward }
   cmpl %esi, %edi
   je .LMMX_SEGMCPY_FINISH      { exit if srcseg:srcofs equals dstseg:dstofs }
   jl .LMMX_SEGMCPY_FORWARD     { or (dstofs < srcofs) then copy forward }

   movl %esi, %edx
   addl %ecx, %edx

   cmpl %edi, %edx              { or ((srcofs+size) < dstofs) then copy forward }
   jl .LMMX_SEGMCPY_FORWARD

.LMMX_SEGMCPY_BACKWARD:
   std                          { prepare for backward copy }

   movl %ecx, %edx              { save size }

   addl %ecx, %esi              { move ofsets to end of block }
   addl %ecx, %edi

   movl %esi, %ebx              { get first number of bytes for alignment to }
                                { source }
   andl $7, %ebx

   cmpl %ebx, %ecx
   jle .LMMX_SEGMCPY_BW_LASTBYTES
   
   movl %ebx, %ecx
   decl %esi; decl %edi
   
   rep
   movsb
   
   incl %esi; incl %edi

   subl %ebx, %edx               { adjust size (subtract first alignment bytes) }
   movl %edx, %ecx

.LMMX_SEGMCPY_BW_QWORD:
   
   shrl $3, %ecx
   jz .LMMX_SEGMCPY_BW_LASTBYTES
.LMMX_SEGMCPY_BW_QWORD_LOOP:
   subl $8, %esi
   subl $8, %edi
   movq (%esi), %mm0
   movq %mm0, %es:(%edi)
   decl %ecx
   jnz .LMMX_SEGMCPY_BW_QWORD_LOOP

   emms                         { clear MMX pipeline }

   movl %edx, %ecx
   
.LMMX_SEGMCPY_BW_LASTBYTES:
   andl $7, %ecx
   decl %esi; decl %edi

   rep
   movsb

   jmp .LMMX_SEGMCPY_FINISH

.LMMX_SEGMCPY_FORWARD:
   cld                          { prepare for forward copy }

   movl %ecx, %edx              { save size }

   movl $8, %ebx
   subl %esi, %ebx              { align on srcseg:srcofs }
   andl $7, %ebx
   jz .LMMX_SEGMCPY_FWD_QWORD   { ebx = number of bytes of first alignment }

   cmpl %ebx, %ecx              { if (size <= ebx) then goto lastbytes }
   jle .LMMX_SEGMCPY_FWD_LASTBYTES

   movl %ebx, %ecx              { prepare to move first alignment bytes }
   rep
   movsb
   
   subl %ebx, %edx
   movl %edx, %ecx              { decrease size by first alignment }
   
.LMMX_SEGMCPY_FWD_QWORD:

   shrl $3, %ecx
   jz .LMMX_SEGMCPY_FWD_LASTBYTES { if (size<8) then goto lastbytes  }

.LMMX_SEGMCPY_FWD_QWORD_LOOP:
   movq (%esi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .LMMX_SEGMCPY_FWD_QWORD_LOOP

   emms                         { clear MMX pipeline }

   movl %edx, %ecx

.LMMX_SEGMCPY_FWD_LASTBYTES:
   andl $7, %ecx                { copy last alignment bytes }

   rep
   movsb

.LMMX_SEGMCPY_FINISH:
   popw %es                     { restore segment registers }
   popw %ds
   
.LMMX_SEGMCPY_EXIT:
end;

procedure seg_memsetb_MMX(seg : Word; ofs : DWord; size : DWord; val : Byte); assembler;
asm
   .ALIGN 4
   { if (size = 0) then exit; }
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .LMMX_SEGMSETB_EXIT

   cld

   { expand byte to dword }
   xorl %eax, %eax
   movb VAL, %al
   movb VAL, %ah
   movl %eax, %ebx
   shll $16, %eax
   orl %ebx, %eax

   { set es:(edi) as target }
   movl OFS, %edi

   movw SEG, %dx
   pushw %es
   movw %dx, %es

   { first_align = (8-address) and $7 }
   movl $8, %ebx
   movl %ecx, %edx
   subl %edi, %ebx
   andl $7, %ebx
   jz .LMMX_SEGMSETB_FILLQWORD

   { if (first_align >= size) then begin}
   cmpl %ebx, %ecx
   jle .LMMX_SEGMSETB_LASTBYTES

   { first_align = number of bytes to plot }
   movl %ebx, %ecx
   shrl %ecx
   jnc .LMMX_SEGMSETB_FA_NOCARRY
   movb %al, %es:(%edi)
   incl %edi
.LMMX_SEGMSETB_FA_NOCARRY:
   
   rep
   stosw

   { edx = edx - first_align }
   subl %ebx, %edx
   movl %edx, %ecx

.LMMX_SEGMSETB_FILLQWORD:

   shrl $3, %ecx
   jz .LMMX_SEGMSETB_SKIPQWORD

   { prepare MMX reg for storing data }
   movd %eax, %mm0
   punpckldq %mm0, %mm0

   { rep stosq }
.LMMX_SEGMSETB_QWORDLOOP:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LMMX_SEGMSETB_QWORDLOOP
   { clear MMX pipeline }
   emms

.LMMX_SEGMSETB_SKIPQWORD:
   { store final bytes }
   andl $7, %edx
   testl %edx, %edx
   jz .LMMX_SEGMSETB_FINISH

   movl %edx, %ecx
   
.LMMX_SEGMSETB_LASTBYTES:
   shrl %ecx
   rep
   stosw
   { carry is preserved by the rep stosw... :) }
   jnc .LMMX_SEGMSETB_LB_NOCARRY
   movb %al, %es:(%edi)
.LMMX_SEGMSETB_LB_NOCARRY:

.LMMX_SEGMSETB_FINISH:

   popw %es

.LMMX_SEGMSETB_EXIT:
end;

procedure seg_memsetw_MMX(seg : Word; ofs : DWord; size : DWord; val : Word); assembler;
asm
   .ALIGN 4
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .LMMX_SEGMSETW_EXIT

   { maximum allowed word count is $7fffffff }
   andl $0x7fffffff, %ecx
   { adjust size to be represented in bytes }
   shll %ecx

   cld

   xorl %eax, %eax
   movw VAL, %ax
   shll $16, %eax
   movw VAL, %ax

   movl OFS, %edi

   movw SEG, %dx
   pushw %es
   movw %dx, %es

   { first_align = (8-address) and $7 }
   movl $8, %ebx
   { save size in bytes to edx }
   movl %ecx, %edx
   subl %edi, %ebx
   andl $7, %ebx
   jz .LMMX_SEGMSETW_FILLQWORD

   { if (first_align >= size) then begin}
   cmpl %ebx, %ecx
   jle .LMMX_SEGMSETW_LASTBYTES

   { first_align = number of bytes to plot }
   movl %ebx, %ecx
   shrl %ecx
   jnc .LMMX_SEGMSETW_FA_NOCARRY
   movb %al, %es:(%edi)
   incl %edi

   rorl $8, %eax
.LMMX_SEGMSETW_FA_NOCARRY:
   
   rep
   stosw

   { edx = edx - first_align }
   subl %ebx, %edx
   movl %edx, %ecx

.LMMX_SEGMSETW_FILLQWORD:

   shrl $3, %ecx
   jz .LMMX_SEGMSETW_SKIPQWORD

   { prepare MMX reg for storing data }
   movd %eax, %mm0
   punpckldq %mm0, %mm0

   { rep stosq }
.LMMX_SEGMSETW_QWORDLOOP:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LMMX_SEGMSETW_QWORDLOOP
   { clear MMX pipeline }
   emms

.LMMX_SEGMSETW_SKIPQWORD:
   { store final bytes }
   andl $7, %edx
   testl %edx, %edx
   jz .LMMX_SEGMSETW_FINISH

   movl %edx, %ecx
   
.LMMX_SEGMSETW_LASTBYTES:
   shrl %ecx
   rep
   stosw
   { carry is preserved by the rep stosw... :) }
   jnc .LMMX_SEGMSETW_LB_NOCARRY
   movb %al, %es:(%edi)
.LMMX_SEGMSETW_LB_NOCARRY:

.LMMX_SEGMSETW_FINISH:

   popw %es
   
.LMMX_SEGMSETW_EXIT:
end;


procedure seg_memsetd_MMX(seg : Word; ofs : DWord; size : DWord; val : DWord); assembler;
asm
   .ALIGN 4
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .LMMX_SEGMSETD_EXIT

   { maximum allowed dword count is $3fffffff }
   andl $0x3fffffff, %ecx
   { adjust size to be represented in bytes }
   shll $2, %ecx
   { store value in eax }
   movl VAL, %eax
   { store offset }
   movl OFS, %edi
   { destination is at %es:(%edi) }
   movw SEG, %dx
   pushw %es
   movw %dx, %es

   { first_align = (8-address) and $7 }
   movl $8, %ebx
   { save size in bytes to edx }
   movl %ecx, %edx
   subl %edi, %ebx
   andl $7, %ebx
   jz .LMMX_SEGMSETD_FILLQWORD

   { if (first_align >= size) then begin}
   cmpl %ebx, %ecx
   jle .LMMX_SEGMSETD_LASTBYTES

   { first_align = number of bytes to plot }
   movl %ebx, %ecx
   andl $3, %ecx
   jz .LMMX_SEGMSETD_FA_NOCARRY

.LMMX_SEGMSETD_FA_LOOP:
   movb %al, %es:(%edi)
   incl %edi
   rorl $8, %eax
   decl %ecx
   jnz .LMMX_SEGMSETD_FA_LOOP
   
.LMMX_SEGMSETD_FA_NOCARRY:

   movl %ebx, %ecx
   testl $0x4, %ecx
   jz .LMMX_SEGMSETD_FA_DCARRY

   movl %eax, %es:(%edi)
   addl $4, %edi
   
.LMMX_SEGMSETD_FA_DCARRY:

   { edx = edx - first_align }
   subl %ebx, %edx
   movl %edx, %ecx

.LMMX_SEGMSETD_FILLQWORD:

   shrl $3, %ecx
   jz .LMMX_SEGMSETD_SKIPQWORD

   { prepare MMX reg for storing data }
   movd %eax, %mm0
   punpckldq %mm0, %mm0

   { rep stosq }
.LMMX_SEGMSETD_QWORDLOOP:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LMMX_SEGMSETD_QWORDLOOP
   { clear MMX pipeline }
   emms

.LMMX_SEGMSETD_SKIPQWORD:
   { store final bytes }
   andl $7, %edx
   testl %edx, %edx
   jz .LMMX_SEGMSETD_FINISH

   movl %edx, %ecx
   
.LMMX_SEGMSETD_LASTBYTES:
   testl $0x4, %ecx
   jz .LMMX_SEGMSETD_LB_NOCARRY

   movl %eax, %es:(%edi)
   addl $4, %edi

.LMMX_SEGMSETD_LB_NOCARRY:
   andl $3,%ecx
   jz .LMMX_SEGMSETB_LB_NOLOOP

.LMMX_SEGMSETD_LB_LOOP:
   movb %al, %es:(%edi)
   rorl $8, %eax
   incl %edi
   decl %ecx
   jnz .LMMX_SEGMSETD_LB_LOOP
   
.LMMX_SEGMSETB_LB_NOLOOP:

.LMMX_SEGMSETD_FINISH:

   popw %es
   
.LMMX_SEGMSETD_EXIT:
end;

{ -------------------------------------------------------------------------- }
{ KNI versions of :                                                          }
{                  seg_memcpy()   (last rev 24.03.1999)                      }
{ -------------------------------------------------------------------------- }

{ copies size bytes of memory between src and dst; KNI version }
procedure seg_memcpy_KNI(srcseg : Word; srcofs : DWord; dstseg : Word; dstofs : DWord; size : DWord); assembler;
asm
   .ALIGN 4
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .LKNI_SEGMCPY_EXIT        { exit if size = 0 }

   movw SRCSEG, %ax
   movw DSTSEG, %bx
   movl SRCOFS, %esi
   movl DSTOFS, %edi

   pushw %ds                    { prepare segment registers }
   pushw %es
   movw %ax, %ds
   movw %bx, %es

   cmpw %ax, %bx
   jne .LKNI_SEGMCPY_FORWARD    { if (srcseg <> dstseg) then copy forward }
   cmpl %esi, %edi
   je .LKNI_SEGMCPY_FINISH      { exit if srcseg:srcofs equals dstseg:dstofs }
   jl .LKNI_SEGMCPY_FORWARD     { or (dstofs < srcofs) then copy forward }

   movl %esi, %edx
   addl %ecx, %edx

   cmpl %edi, %edx              { or ((srcofs+size) < dstofs) then copy forward }
   jl .LKNI_SEGMCPY_FORWARD

.LKNI_SEGMCPY_BACKWARD:
   std                          { prepare for backward copy }

   movl %ecx, %edx              { save size }

   addl %ecx, %esi              { move offsets to end of block }
   addl %ecx, %edi

   movl %esi, %ebx              { get first number of bytes for alignment to }
                                { source }
   andl $15, %ebx

   cmpl %ebx, %ecx
   jle .LKNI_SEGMCPY_BW_LASTBYTES
   
   movl %ebx, %ecx
   decl %esi; decl %edi
   
   rep
   movsb
   
   incl %esi; incl %edi

   subl %ebx, %edx               { adjust size (subtract first alignment bytes) }
   movl %edx, %ecx

.LKNI_SEGMCPY_BW_QWORD:
   
   shrl $4, %ecx
   jz .LKNI_SEGMCPY_BW_LASTBYTES

.LKNI_SEGMCPY_BW_QWORD_LOOP16:
   decl %ecx
   js .LKNI_SEGMCPY_BW_NOLOOP16
   subl $16, %esi
   subl $16, %edi
   // movaps (%esi), %xmm0
   .byte 0x26, 0x67, 0x0f, 0x28, 0x07
   // movups %xmm0, %es:(%edi)
   .byte 0x3e, 0x67, 0x0f, 0x11, 0x06
   jmp .LKNI_SEGMCPY_BW_QWORD_LOOP16

.LKNI_SEGMCPY_BW_NOLOOP16:
   movl %edx, %ecx
   
.LKNI_SEGMCPY_BW_LASTBYTES:
   andl $15, %ecx
   decl %esi; decl %edi

   rep
   movsb

   jmp .LKNI_SEGMCPY_FINISH

.LKNI_SEGMCPY_FORWARD:
   cld                          { prepare for forward copy }

   movl %ecx, %edx              { save size }

   movl $16, %ebx
   subl %esi, %ebx              { align on srcseg:srcofs }
   andl $15, %ebx
   jz .LKNI_SEGMCPY_FWD_QWORD   { ebx = number of bytes of first alignment }

   cmpl %ebx, %ecx              { if (size <= ebx) then goto lastbytes }
   jle .LKNI_SEGMCPY_FWD_LASTBYTES

   movl %ebx, %ecx              { prepare to move first alignment bytes }
   rep
   movsb
   
   subl %ebx, %edx
   movl %edx, %ecx              { decrease size by first alignment }
   
.LKNI_SEGMCPY_FWD_QWORD:

   shrl $4, %ecx
   jz .LKNI_SEGMCPY_FWD_LASTBYTES { if (size<16) then goto lastbytes  }

.LKNI_SEGMCPY_FWD_QWORD_LOOP16:
   decl %ecx
   js .LKNI_SEGMCPY_FWD_NOLOOP16
   // movaps (%esi), %xmm0
   .byte 0x26, 0x67, 0x0f, 0x28, 0x07
   // movups %xmm0, %es:(%edi)
   .byte 0x3e, 0x67, 0x0f, 0x11, 0x06
   addl $16, %esi
   addl $16, %edi
   jmp .LKNI_SEGMCPY_FWD_QWORD_LOOP16

.LKNI_SEGMCPY_FWD_NOLOOP16:

   movl %edx, %ecx

.LKNI_SEGMCPY_FWD_LASTBYTES:
   andl $15, %ecx                { copy last alignment bytes }

   rep
   movsb

.LKNI_SEGMCPY_FINISH:
   popw %es                     { restore segment registers }
   popw %ds
   
.LKNI_SEGMCPY_EXIT:
end;


{ -------------------------------------------------------------------------- }
{ STD versions of :                                                          }
{                  seg_memcpy()   (last rev 02.01.1999)                      }
{                  seg_memset()   (last rev 02.01.1999)                      }
{                  seg_memsetw()  (last rev 02.01.1999)                      }
{                  seg_memsetd()  (last rev 02.01.1999)                      }
{ -------------------------------------------------------------------------- }

procedure seg_memcpy_std(srcseg : Word; srcofs : DWord; dstseg : Word; dstofs : DWord; size : DWord); assembler;
asm
   .ALIGN 4
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .LSTD_SEGMCPY_EXIT        { exit if size = 0 }

   movw SRCSEG, %ax
   movw DSTSEG, %bx
   movl SRCOFS, %esi
   movl DSTOFS, %edi

   pushw %ds                    { prepare segment registers }
   pushw %es
   movw %ax, %ds
   movw %bx, %es

   cmpw %ax, %bx
   jne .LSTD_SEGMCPY_FORWARD    { if (srcseg <> dstseg) then copy forward }
   cmpl %esi, %edi
   je .LSTD_SEGMCPY_FINISH      { exit if srcseg:srcofs equals dstseg:dstofs }
   jl .LSTD_SEGMCPY_FORWARD     { or (dstofs < srcofs) then copy forward }

   movl %esi, %edx
   addl %ecx, %edx

   cmpl %edi, %edx              { or ((srcofs+size) < dstofs) then copy forward }
   jl .LSTD_SEGMCPY_FORWARD

.LSTD_SEGMCPY_BACKWARD:
   std                          { prepare for backward copy }

   movl %ecx, %edx              { save size }

   addl %ecx, %esi              { move ofsets to end of block }
   addl %ecx, %edi

   movl %esi, %ebx              { get first number of bytes for alignment to }
                                { source }
   andl $3, %ebx

   cmpl %ebx, %ecx
   jle .LSTD_SEGMCPY_BW_LASTBYTES
   
   movl %ebx, %ecx
   decl %esi; decl %edi
   
   rep
   movsb
   
   incl %esi; incl %edi

   subl %ebx, %edx               { adjust size (subtract first alignment bytes) }
   movl %edx, %ecx

.LSTD_SEGMCPY_BW_DWORD:
   
   shrl $2, %ecx
   jz .LSTD_SEGMCPY_BW_LASTBYTES

   subl $4, %esi; subl $4, %edi

   rep
   movsl

   addl $4, %esi; addl $4, %edi

   movl %edx, %ecx
   
.LSTD_SEGMCPY_BW_LASTBYTES:
   andl $3, %ecx
   decl %esi; decl %edi

   rep
   movsb

   jmp .LSTD_SEGMCPY_FINISH

.LSTD_SEGMCPY_FORWARD:
   cld                          { prepare for forward copy }

   movl %ecx, %edx              { save size }

   movl $4, %ebx
   subl %esi, %ebx              { align on srcseg:srcofs }
   andl $3, %ebx
   jz .LSTD_SEGMCPY_FWD_QWORD   { ebx = number of bytes of first alignment }

   cmpl %ebx, %ecx              { if (size <= ebx) then goto lastbytes }
   jle .LSTD_SEGMCPY_FWD_LASTBYTES

   movl %ebx, %ecx              { prepare to move first alignment bytes }
   rep
   movsb
   
   subl %ebx, %edx
   movl %edx, %ecx              { decrease size by first alignment }
   
.LSTD_SEGMCPY_FWD_QWORD:

   shrl $2, %ecx
   jz .LSTD_SEGMCPY_FWD_LASTBYTES { if (size<8) then goto lastbytes  }

   rep
   movsl

   movl %edx, %ecx

.LSTD_SEGMCPY_FWD_LASTBYTES:
   andl $3, %ecx                { copy last alignment bytes }

   rep
   movsb

.LSTD_SEGMCPY_FINISH:
   popw %es                     { restore segment registers }
   popw %ds
   
.LSTD_SEGMCPY_EXIT:
end;

procedure seg_memsetb_STD(seg : Word; ofs : DWord; size : DWord; val : Byte); assembler;
asm
   .ALIGN 4
   { if (size = 0) then exit; }
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .LSTD_SEGMSETB_EXIT

   cld

   { expand byte to dword }
   xorl %eax, %eax
   movb VAL, %al
   movb VAL, %ah
   movl %eax, %ebx
   shll $16, %eax
   orl %ebx, %eax

   { set es:(edi) as target }
   movl OFS, %edi

   movw SEG, %dx
   pushw %es
   movw %dx, %es

   { first_align = (4-address) and $3 }
   movl $4, %ebx
   movl %ecx, %edx
   subl %edi, %ebx
   andl $3, %ebx
   jz .LSTD_SEGMSETB_FILLDWORD

   { if (first_align >= size) then begin}
   cmpl %ebx, %ecx
   jle .LSTD_SEGMSETB_LASTBYTES

   { first_align = number of bytes to plot }
   movl %ebx, %ecx
   rep
   stosb

   { edx = edx - first_align }
   subl %ebx, %edx
   movl %edx, %ecx

.LSTD_SEGMSETB_FILLDWORD:

   shrl $2, %ecx
   jz .LSTD_SEGMSETB_SKIPDWORD

   rep
   stosl

.LSTD_SEGMSETB_SKIPDWORD:
   { store final bytes }
   andl $3, %edx
   testl %edx, %edx
   jz .LSTD_SEGMSETB_FINISH

   movl %edx, %ecx
   
.LSTD_SEGMSETB_LASTBYTES:
   rep
   stosb

.LSTD_SEGMSETB_FINISH:

   popw %es
.LSTD_SEGMSETB_EXIT:
end;

procedure seg_memsetw_std(seg : Word; ofs : DWord; size : DWord; val : Word); assembler;
asm
   .ALIGN 4
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .LSTD_SEGMSETW_EXIT

   { maximum allowed word count is $7fffffff }
   andl $0x7fffffff, %ecx
   { adjust size to be represented in bytes }
   shll %ecx

   cld

   xorl %eax, %eax
   movw VAL, %ax
   shll $16, %eax
   movw VAL, %ax

   movl OFS, %edi

   movw SEG, %dx
   pushw %es
   movw %dx, %es

   { first_align = (4-address) and $3 }
   movl $4, %ebx
   { save size in bytes to edx }
   movl %ecx, %edx
   subl %edi, %ebx
   andl $3, %ebx
   jz .LSTD_SEGMSETW_FILLDWORD

   { if (first_align >= size) then begin}
   cmpl %ebx, %ecx
   jle .LSTD_SEGMSETW_LASTBYTES

   { first_align = number of bytes to plot }
   movl %ebx, %ecx

.LSTD_SEGMSETW_FA_LOOP:
   movb %al, %es:(%edi)
   incl %edi
   rorl $8, %eax
   decl %ecx
   jnz .LSTD_SEGMSETW_FA_LOOP
   
   { edx = edx - first_align }
   subl %ebx, %edx
   movl %edx, %ecx

.LSTD_SEGMSETW_FILLDWORD:

   shrl $2, %ecx
   jz .LSTD_SEGMSETW_SKIPDWORD

   rep
   stosl

.LSTD_SEGMSETW_SKIPDWORD:
   { store final bytes }
   andl $3, %edx
   testl %edx, %edx
   jz .LSTD_SEGMSETW_FINISH

   movl %edx, %ecx
   
.LSTD_SEGMSETW_LASTBYTES:

.LSTD_SEGMSETW_LB_LOOP:
   movb %al, %es:(%edi)
   incl %edi
   rorl $8, %eax
   decl %ecx
   jnz .LSTD_SEGMSETW_LB_LOOP
   
.LSTD_SEGMSETW_FINISH:

   popw %es
   
.LSTD_SEGMSETW_EXIT:
end;

procedure seg_memsetd_STD(seg : Word; ofs : DWord; size : DWord; val : DWord); assembler;
asm
   .ALIGN 4
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .LSTD_SEGMSETD_EXIT

   { maximum allowed dword count is $3fffffff }
   andl $0x3fffffff, %ecx
   { adjust size to be represented in bytes }
   shll $2, %ecx
   { store value in eax }
   movl VAL, %eax
   { store offset }
   movl OFS, %edi
   { destination is at %es:(%edi) }
   movw SEG, %dx
   pushw %es
   movw %dx, %es

   { first_align = (4-address) and $3 }
   movl $4, %ebx
   { save size in bytes to edx }
   movl %ecx, %edx
   subl %edi, %ebx
   andl $3, %ebx
   jz .LSTD_SEGMSETD_FILLDWORD

   { if (first_align >= size) then begin}
   cmpl %ebx, %ecx
   jle .LSTD_SEGMSETD_LASTBYTES

   { first_align = number of bytes to plot }
   movl %ebx, %ecx
   andl $3, %ecx

.LSTD_SEGMSETD_FA_LOOP:
   movb %al, %es:(%edi)
   incl %edi
   rorl $8, %eax
   decl %ecx
   jnz .LSTD_SEGMSETD_FA_LOOP
   
.LSTD_SEGMSETD_FA_NOCARRY:

   { edx = edx - first_align }
   subl %ebx, %edx
   movl %edx, %ecx

.LSTD_SEGMSETD_FILLDWORD:

   shrl $2, %ecx
   jz .LSTD_SEGMSETD_SKIPDWORD

   rep
   stosl

.LSTD_SEGMSETD_SKIPDWORD:
   { store final bytes }
   andl $3, %edx
   testl %edx, %edx
   jz .LSTD_SEGMSETD_FINISH

   movl %edx, %ecx
   
.LSTD_SEGMSETD_LASTBYTES:

.LSTD_SEGMSETD_LB_LOOP:
   movb %al, %es:(%edi)
   rorl $8, %eax
   incl %edi
   decl %ecx
   jnz .LSTD_SEGMSETD_LB_LOOP
   
.LSTD_SEGMSETD_FINISH:

   popw %es
   
.LSTD_SEGMSETD_EXIT:
end;

{$ASMMODE DIRECT}

{ -------------------------------------------------------------------------- }
{ MMX versions of :                                                          }
{                  seg_memchange()          (last rev. 15.10.1998)           }
{                  seg_memchangeValueB()    (last rev. 15.10.1998)           }
{                  seg_memchangeValueW()    (last rev. 15.10.1998)           }
{                  seg_memchangeValueD()    (last rev. 15.10.1998)           }
{ -------------------------------------------------------------------------- }

{ does a boolean operation between two memory areas; MMX version }
procedure seg_memchangeMMX (srcsel : Word; srcofs : DWord; dstsel : Word; dstofs : DWord; size : DWord; op : Mem_Op); assembler;
asm
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .Lseg_memchangeMMXExit  { if (size = 0) then exit; }
   movw SRCSEL, %bx
   movw DSTSEL, %dx
   movl SRCOFS, %esi
   movl DSTOFS, %edi
   movl OP, %eax
   pushw %ds
   pushw %es
   movw %bx, %ds           { src = ds:(esi) }
   movw %dx, %es           { dst = es:(edi) }

   cld                     { clear direction flag for forward copy }

   movl %eax, %ebx
   movl %ecx, %edx         { save size }
   andl $7, %ecx
   jz .Lseg_memchangeMMXQuad { if (size and $7) <> 0 then }
   jmpl .Lseg_memchangeMMXByteTable(, %ebx, 4) { case (op) of }
.Lseg_memchangeMMXByteTable:
.long .Lseg_memchangeMMXByteCLEAR          { dst := 0 }
.long .Lseg_memchangeMMXByteNOT_DSTORSRC   { dst := not (dst or src) }
.long .Lseg_memchangeMMXByteDST_AND_NOTSRC { dst := dst and (not src) }
.long .Lseg_memchangeMMXByteNOTSRC         { dst := not src }
.long .Lseg_memchangeMMXByteSRC_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeMMXByteNOTDST         { dst := not dst }
.long .Lseg_memchangeMMXByteDST_XOR_SRC    { dst := dst xor src }
.long .Lseg_memchangeMMXByteNOT_DSTANDSRC  { dst := not (dst and src) }
.long .Lseg_memchangeMMXByteDST_AND_SRC    { dst := dst and src }
.long .Lseg_memchangeMMXByteNOT_DSTXORSRC  { dst := not (dst xor src) }
.long .Lseg_memchangeMMXByteDST            { dst := dst; do nothing }
.long .Lseg_memchangeMMXByteDST_OR_NOTSRC  { dst := dst or (not src) }
.long .Lseg_memchangeMMXByteSRC            { dst := src; copy }
.long .Lseg_memchangeMMXByteNOTDST_OR_SRC  { dst := (not dst) or src }
.long .Lseg_memchangeMMXByteDST_OR_SRC     { dst := dst or src }
.long .Lseg_memchangeMMXByteSET            { dst := 1 }

.Lseg_memchangeMMXByteCLEAR:          { dst := 0 }
   xorb %al, %al
   rep
   stosb
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteNOT_DSTORSRC:   { dst := not (dst or src) }
   movb (%esi), %al
   orb  %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXByteNOT_DSTORSRC
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteDST_AND_NOTSRC: { dst := dst and (not src) }
   movb (%esi), %al
   notb %al
   andb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXByteDST_AND_NOTSRC
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteNOTSRC:         { dst := not src }
   movb (%esi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXByteNOTSRC
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteSRC_AND_NOTDST: { dst := src and (not dst) }
   movb %es:(%edi), %al
   notb %al
   andb (%esi), %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXByteSRC_AND_NOTDST
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteNOTDST:         { dst := not dst }
   notb %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXByteNOTDST
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteDST_XOR_SRC:    { dst := dst xor src }
   movb (%esi), %al
   xorb %al, %es:(%edi)
   incl %edi
   incl %esi
   decl %ecx
   jnz .Lseg_memchangeMMXByteDST_XOR_SRC
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteNOT_DSTANDSRC:  { dst := not (dst and src) }
   movb (%esi), %al
   andb %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXByteNOT_DSTANDSRC
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteDST_AND_SRC:    { dst := dst and src }
   movb (%esi), %al
   andb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXByteDST_AND_SRC
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteNOT_DSTXORSRC:  { dst := not (dst xor src) }
   movb (%esi), %al
   xorb %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXByteNOT_DSTXORSRC
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXByteDST_OR_NOTSRC:  { dst := dst or (not src) }
   movb (%esi), %al
   notb %al
   orb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXByteDST_OR_NOTSRC
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteSRC:            { dst := src; copy }
   rep
   movsb
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteNOTDST_OR_SRC:  { dst := (not dst) or src }
   movb %es:(%edi), %al
   notb %al
   orb (%esi), %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXByteNOTDST_OR_SRC
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteDST_OR_SRC:     { dst := dst or src }
   movb (%esi), %al
   orb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXByteDST_OR_SRC
   jmp .Lseg_memchangeMMXQuad
.Lseg_memchangeMMXByteSET:            { dst := 1 }
   movl $0xFFFFFFFF, %eax
   rep
   stosb

.Lseg_memchangeMMXQuad:
   movl %edx, %ecx            { ecx = size div 8 }
   shrl $3, %ecx
   jz .Lseg_memchangeNoMMX
   jmpl .Lseg_memchangeMMXQuadTable(, %ebx, 4)
.Lseg_memchangeMMXQuadTable:
.long .Lseg_memchangeMMXQuadCLEAR          { dst := 0 }
.long .Lseg_memchangeMMXQuadNOT_DSTORSRC   { dst := not (dst or src) }
.long .Lseg_memchangeMMXQuadDST_AND_NOTSRC { dst := dst and (not src) }
.long .Lseg_memchangeMMXQuadNOTSRC         { dst := not src }
.long .Lseg_memchangeMMXQuadSRC_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeMMXQuadNOTDST         { dst := not dst }
.long .Lseg_memchangeMMXQuadDST_XOR_SRC    { dst := dst xor src }
.long .Lseg_memchangeMMXQuadNOT_DSTANDSRC  { dst := not (dst and src) }
.long .Lseg_memchangeMMXQuadDST_AND_SRC    { dst := dst and src }
.long .Lseg_memchangeMMXQuadNOT_DSTXORSRC  { dst := not (dst xor src) }
.long .Lseg_memchangeMMXQuadDST            { dst := dst; do nothing }
.long .Lseg_memchangeMMXQuadDST_OR_NOTSRC  { dst := dst or (not src) }
.long .Lseg_memchangeMMXQuadSRC            { dst := src; copy }
.long .Lseg_memchangeMMXQuadNOTDST_OR_SRC  { dst := (not dst) or src }
.long .Lseg_memchangeMMXQuadDST_OR_SRC     { dst := dst or src }
.long .Lseg_memchangeMMXQuadSET            { dst := 1 }

.Lseg_memchangeMMXQuadCLEAR:          { dst := 0 }
   movq MMXQuadClear, %mm0
.LMMXQuadClear:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LMMXQuadClear
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadNOT_DSTORSRC:   { dst := not (dst or src) }
   movq (%esi), %mm0
   por  %es:(%edi), %mm0
   pandn MMXQuadSet, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadNOT_DSTORSRC
   jmp .Lseg_memchangeMMXFinish

.Lseg_memchangeMMXQuadDST_AND_NOTSRC: { dst := dst and (not src) }
   movq (%esi), %mm0
   pandn %es:(%edi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadDST_AND_NOTSRC
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadNOTSRC:         { dst := not src }
   movq MMXQuadSet, %mm0
   pxor (%esi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadNOTSRC
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadSRC_AND_NOTDST: { dst := src and (not dst) }
   movq %es:(%edi), %mm0
   pandn (%esi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadSRC_AND_NOTDST
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadNOTDST:         { dst := not dst }
   movq MMXQuadSet, %mm0
   pxor %es:(%edi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadNOTDST
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadDST_XOR_SRC:    { dst := dst xor src }
   movq %es:(%edi), %mm0
   pxor (%esi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadDST_XOR_SRC
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadNOT_DSTANDSRC:  { dst := not (dst and src) }
   movq %es:(%edi), %mm0
   pand (%esi), %mm0
   pandn MMXQuadSet, %mm0             { %mm0 := not %mm0 }
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadNOT_DSTANDSRC
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadDST_AND_SRC:    { dst := dst and src }
   movq %es:(%edi), %mm0
   pand (%esi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadDST_AND_SRC
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadNOT_DSTXORSRC:  { dst := not (dst xor src) }
   movq %es:(%edi), %mm0
   pxor (%esi), %mm0
   pandn MMXQuadSet, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadNOT_DSTXORSRC
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadDST_OR_NOTSRC:  { dst := dst or (not src) }
   movq (%esi), %mm0
   pandn MMXQuadSet, %mm0
   por %es:(%edi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadDST_OR_NOTSRC
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadSRC:            { dst := src; copy }
   movq (%esi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadSRC
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadNOTDST_OR_SRC:  { dst := (not dst) or src }
   movq %es:(%edi), %mm0
   pandn MMXQuadSet, %mm0
   por (%esi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadNOTDST_OR_SRC
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadDST_OR_SRC:     { dst := dst or src }
   movq %es:(%edi), %mm0
   por (%esi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %esi
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXQuadDST_OR_SRC
   jmp .Lseg_memchangeMMXFinish
.Lseg_memchangeMMXQuadSET:            { dst := 1 }
   movq MMXQuadSet, %mm0
.LMMXQuadSET:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LMMXQuadSET

.Lseg_memchangeMMXFinish:
   emms
.Lseg_memchangeNoMMX:
   popw %es
   popw %ds

.Lseg_memchangeMMXExit:
   leave
   ret $20
MMXQuadClear:
   .quad 0x0000000000000000
MMXQuadSet:
   .quad 0xffffffffffffffff
end;

{ does a boolean operation between memory and a byte-sized value; MMX version }
procedure seg_memchangeValueBMMX (sel : Word; ofs : DWord; size : DWord; value : Byte; op : Mem_Op); assembler;
asm
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .Lseg_memchangeMMXValueExit  { if (size = 0) then exit; }
   movw SEL, %bx
   movl OFS, %edi
   pushw %es
   movw %bx, %es           { dst = es:(edi) }
   { expand char to MMX quad }
   movb VALUE, %al
   movb %al, %ah
   movw %ax, %bx
   roll $16, %eax
   movw %bx, %ax
   movl %eax, MMXQuadVarC
   movl %eax, 4+MMXQuadVarC

   cld                     { clear direction flag for forward copy }

   movl OP, %ebx
   movl %ecx, %edx         { save size }
   andl $7, %ecx
   jz .Lseg_memchangeMMXValueQuad { if (size and $7) <> 0 then }
   jmpl .Lseg_memchangeMMXValueByteTable(, %ebx, 4) { case (op) of }
.Lseg_memchangeMMXValueByteTable:
.long .Lseg_memchangeMMXValueByteCLEAR          { dst := 0 }
.long .Lseg_memchangeMMXValueByteNOT_DSTORVAL   { dst := not (dst or src) }
.long .Lseg_memchangeMMXValueByteDST_AND_NOTVAL { dst := dst and (not src) }
.long .Lseg_memchangeMMXValueByteNOTVAL         { dst := not src }
.long .Lseg_memchangeMMXValueByteVAL_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeMMXValueByteNOTDST         { dst := not dst }
.long .Lseg_memchangeMMXValueByteDST_XOR_VAL    { dst := dst xor src }
.long .Lseg_memchangeMMXValueByteNOT_DSTANDVAL  { dst := not (dst and src) }
.long .Lseg_memchangeMMXValueByteDST_AND_VAL    { dst := dst and src }
.long .Lseg_memchangeMMXValueByteNOT_DSTXORVAL  { dst := not (dst xor src) }
.long .Lseg_memchangeMMXValueByteDST            { dst := dst; do nothing }
.long .Lseg_memchangeMMXValueByteDST_OR_NOTVAL  { dst := dst or (not src) }
.long .Lseg_memchangeMMXValueByteVAL            { dst := src; copy }
.long .Lseg_memchangeMMXValueByteNOTDST_OR_VAL  { dst := (not dst) or src }
.long .Lseg_memchangeMMXValueByteDST_OR_VAL     { dst := dst or src }
.long .Lseg_memchangeMMXValueByteSET            { dst := 1 }

.Lseg_memchangeMMXValueByteCLEAR:          { dst := 0 }
   xorb %al, %al
   rep
   stosb
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteNOT_DSTORVAL:   { dst := not (dst or value) }
   movb %ah, %al
   orb  %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueByteNOT_DSTORVAL
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteDST_AND_NOTVAL: { dst := dst and (not value) }
   notb %al
.LMMXValueByteDST_AND_NOTVAL:
   andb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .LMMXValueByteDST_AND_NOTVAL
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteNOTVAL:         { dst := not value }
   notb %al
   rep
   stosb
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteVAL_AND_NOTDST: { dst := value and (not dst) }
   movb %es:(%edi), %al
   notb %al
   andb %ah, %al
   movb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueByteVAL_AND_NOTDST
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteNOTDST:         { dst := not dst }
   notb %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueByteNOTDST
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteDST_XOR_VAL:    { dst := dst xor value }
   xorb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueByteDST_XOR_VAL
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteNOT_DSTANDVAL:  { dst := not (dst and value) }
   movb %ah, %al
   andb %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueByteNOT_DSTANDVAL
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteDST_AND_VAL:    { dst := dst and value }
   andb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueByteDST_AND_VAL
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteNOT_DSTXORVAL:  { dst := not (dst xor value) }
   movb %ah, %al
   xorb %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueByteNOT_DSTXORVAL
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueByteDST_OR_NOTVAL:  { dst := dst or (not value) }
   notb %al
.LMMXValueByteDST_OR_NOTVAL:
   orb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .LMMXValueByteDST_OR_NOTVAL
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteVAL:            { dst := value; fill }
   rep
   stosb
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteNOTDST_OR_VAL:  { dst := (not dst) or value }
   movb %es:(%edi), %al
   notb %al
   orb %ah, %al
   movb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueByteNOTDST_OR_VAL
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteDST_OR_VAL:     { dst := dst or value }
   orb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueByteDST_OR_VAL
   jmp .Lseg_memchangeMMXValueQuad
.Lseg_memchangeMMXValueByteSET:            { dst := 1 }
   movb $0xFF, %al
   rep
   stosb

.Lseg_memchangeMMXValueQuad:
   movl %esi, %eax            { restore eax }
   movl %edx, %ecx            { ecx = size div 8 }
   shrl $3, %ecx
   jz .Lseg_memchangeMMXValueFinish
   jmpl .Lseg_memchangeMMXValueQuadTable(, %ebx, 4)
.Lseg_memchangeMMXValueQuadTable:
.long .Lseg_memchangeMMXValueQuadCLEAR          { dst := 0 }
.long .Lseg_memchangeMMXValueQuadNOT_DSTORVAL   { dst := not (dst or src) }
.long .Lseg_memchangeMMXValueQuadDST_AND_NOTVAL { dst := dst and (not src) }
.long .Lseg_memchangeMMXValueQuadNOTVAL         { dst := not src }
.long .Lseg_memchangeMMXValueQuadVAL_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeMMXValueQuadNOTDST         { dst := not dst }
.long .Lseg_memchangeMMXValueQuadDST_XOR_VAL    { dst := dst xor src }
.long .Lseg_memchangeMMXValueQuadNOT_DSTANDVAL  { dst := not (dst and src) }
.long .Lseg_memchangeMMXValueQuadDST_AND_VAL    { dst := dst and src }
.long .Lseg_memchangeMMXValueQuadNOT_DSTXORVAL  { dst := not (dst xor src) }
.long .Lseg_memchangeMMXValueQuadDST            { dst := dst; do nothing }
.long .Lseg_memchangeMMXValueQuadDST_OR_NOTVAL  { dst := dst or (not src) }
.long .Lseg_memchangeMMXValueQuadVAL            { dst := src; copy }
.long .Lseg_memchangeMMXValueQuadNOTDST_OR_VAL  { dst := (not dst) or src }
.long .Lseg_memchangeMMXValueQuadDST_OR_VAL     { dst := dst or src }
.long .Lseg_memchangeMMXValueQuadSET            { dst := 1 }

.Lseg_memchangeMMXValueQuadCLEAR:          { dst := 0 }
   movq MMXQuadClearC, %mm0
.LValueQuadCLEAR:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueQuadCLEAR
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadNOT_DSTORVAL:   { dst := not (dst or value) }
   movq MMXQuadVarC, %mm0
   por  %es:(%edi), %mm0
   pandn MMXQuadSetC, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueQuadNOT_DSTORVAL
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadDST_AND_NOTVAL: { dst := dst and (not value) }
   movq MMXQuadVarC, %mm0
   pandn MMXQuadSetC, %mm0
   movq %mm0, MMXQuadVarC
.LValueQuadDST_AND_NOTVAL:
   movq MMXQuadVarC, %mm0
   pand %es:(%edi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueQuadDST_AND_NOTVAL
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadNOTVAL:         { dst := not value }
   movq MMXQuadVarC, %mm0
   pandn MMXQuadSetC, %mm0
.LValueQuadNOTVAL:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueQuadNOTVAL
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadVAL_AND_NOTDST: { dst := value and (not dst) }
   movq %es:(%edi), %mm0
   pandn MMXQuadSetC, %mm0
   pand MMXQuadVarC, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueQuadVAL_AND_NOTDST
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadNOTDST:         { dst := not dst }
   movq %es:(%edi), %mm0
   pandn MMXQuadSetC, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueQuadNOTDST
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadDST_XOR_VAL:    { dst := dst xor value }
   movq %es:(%edi), %mm0
   pxor MMXQuadVarC, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueQuadDST_XOR_VAL
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadNOT_DSTANDVAL:  { dst := not (dst and value) }
   movq %es:(%edi), %mm0
   pandl MMXQuadVarC, %mm0
   pandn MMXQuadSetC, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueQuadNOT_DSTANDVAL
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadDST_AND_VAL:    { dst := dst and value }
   movq %es:(%edi), %mm0
   pand MMXQuadVarC, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueQuadDST_AND_VAL
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadNOT_DSTXORVAL:  { dst := not (dst xor value) }
   movq %es:(%edi), %mm0
   pxor MMXQuadVarC, %mm0
   pandn MMXQuadSetC, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueQuadNOT_DSTXORVAL
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadDST_OR_NOTVAL:  { dst := dst or (not value) }
   movq MMXQuadVarC, %mm0
   pandn MMXQuadSetC, %mm0
.LValueQuadDST_OR_NOTVAL:
   movq %mm0, %mm1
   por %es:(%edi), %mm1
   movq %mm1, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueQuadDST_OR_NOTVAL
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadVAL:            { dst := value; fill }
   movq MMXQuadVarC, %mm0
.LValueQuadVAL:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueQuadVAL
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadNOTDST_OR_VAL:  { dst := (not dst) or value }
   movq %es:(%edi), %mm0
   pandn MMXQuadSetC, %mm0
   por MMXQuadVarC, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueQuadNOTDST_OR_VAL
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadDST_OR_VAL:     { dst := dst or value }
   movq %es:(%edi), %mm0
   por MMXQuadVarC, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueQuadDST_OR_VAL
   jmp .Lseg_memchangeMMXValueFinish
.Lseg_memchangeMMXValueQuadSET:            { dst := 1 }
   movq MMXQuadSetC, %mm0
.LValueQuadSET:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueQuadSET

.Lseg_memchangeMMXValueFinish:
   popw %es
   emms
.Lseg_memchangeMMXValueExit:
   leave
   ret $16
MMXQuadSetC:
   .quad 0xffffffffffffffff { MMX reg with all bits set }
MMXQuadClearC:
   .quad 0x0000000000000000 { MMX reg with all bits clear }
MMXQuadVarC:
   .quad 0x1234567890abcdef { MMX reg to store value in }
end;

{ does a boolean operation between memory and a word-sized value; MMX version }
procedure seg_memchangeValueWMMX (sel : Word; ofs : DWord; size : DWord; value : Word; op : Mem_Op); assembler;
asm
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .Lseg_memchangeMMXValueWExit  { if (size = 0) then exit; }
   movw SEL, %bx
   movl OFS, %edi
   pushw %es
   movw %bx, %es           { dst = es:(edi) }
   { expand word to MMX quad }
   movw Value, %ax
   roll $16, %eax
   movw VALUE, %ax
   movl %eax, MMXQuadVarW
   movl %eax, 4+MMXQuadVarW

   cld                     { clear direction flag for forward copy }

   movl %eax, %esi
   movl OP, %ebx
   movl %ecx, %edx         { save size }
   andl $3, %ecx
   jz .Lseg_memchangeMMXValueWQuad { if (size and $3) <> 0 then }
   jmpl .Lseg_memchangeMMXValueWByteTable(, %ebx, 4) { case (op) of }
.Lseg_memchangeMMXValueWByteTable:
.long .Lseg_memchangeMMXValueWByteCLEAR          { dst := 0 }
.long .Lseg_memchangeMMXValueWByteNOT_DSTORVAL   { dst := not (dst or src) }
.long .Lseg_memchangeMMXValueWByteDST_AND_NOTVAL { dst := dst and (not src) }
.long .Lseg_memchangeMMXValueWByteNOTVAL         { dst := not src }
.long .Lseg_memchangeMMXValueWByteVAL_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeMMXValueWByteNOTDST         { dst := not dst }
.long .Lseg_memchangeMMXValueWByteDST_XOR_VAL    { dst := dst xor src }
.long .Lseg_memchangeMMXValueWByteNOT_DSTANDVAL  { dst := not (dst and src) }
.long .Lseg_memchangeMMXValueWByteDST_AND_VAL    { dst := dst and src }
.long .Lseg_memchangeMMXValueWByteNOT_DSTXORVAL  { dst := not (dst xor src) }
.long .Lseg_memchangeMMXValueWByteDST            { dst := dst; do nothing }
.long .Lseg_memchangeMMXValueWByteDST_OR_NOTVAL  { dst := dst or (not src) }
.long .Lseg_memchangeMMXValueWByteVAL            { dst := src; copy }
.long .Lseg_memchangeMMXValueWByteNOTDST_OR_VAL  { dst := (not dst) or src }
.long .Lseg_memchangeMMXValueWByteDST_OR_VAL     { dst := dst or src }
.long .Lseg_memchangeMMXValueWByteSET            { dst := 1 }

.Lseg_memchangeMMXValueWByteCLEAR:          { dst := 0 }
   xorw %ax, %ax
   rep
   stosw
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteNOT_DSTORVAL:   { dst := not (dst or ValueW) }
   movw %si, %ax
   orw  %es:(%edi), %ax
   notw %ax
   movw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWByteNOT_DSTORVAL
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteDST_AND_NOTVAL: { dst := dst and (not ValueW) }
   notw %ax
.LMMXValueWByteDST_AND_NOTVAL:
   andw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .LMMXValueWByteDST_AND_NOTVAL
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteNOTVAL:         { dst := not ValueW }
   notw %ax
   rep
   stosw
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteVAL_AND_NOTDST: { dst := ValueW and (not dst) }
   movw %es:(%edi), %ax
   notw %ax
   andw %si, %ax
   movw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWByteVAL_AND_NOTDST
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteNOTDST:         { dst := not dst }
   notw %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWByteNOTDST
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteDST_XOR_VAL:    { dst := dst xor ValueW }
   xorw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWByteDST_XOR_VAL
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteNOT_DSTANDVAL:  { dst := not (dst and ValueW) }
   movw %si, %ax
   andb %es:(%edi), %ax
   notw %ax
   movw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWByteNOT_DSTANDVAL
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteDST_AND_VAL:    { dst := dst and ValueW }
   andw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWByteDST_AND_VAL
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteNOT_DSTXORVAL:  { dst := not (dst xor ValueW) }
   movw %si, %ax
   xorw %es:(%edi), %ax
   notw %ax
   movw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWByteNOT_DSTXORVAL
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWByteDST_OR_NOTVAL:  { dst := dst or (not ValueW) }
   notw %ax
.LMMXValueWByteDST_OR_NOTVAL:
   orw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .LMMXValueWByteDST_OR_NOTVAL
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteVAL:            { dst := ValueW; fill }
   rep
   stosw
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteNOTDST_OR_VAL:  { dst := (not dst) or ValueW }
   movw %es:(%edi), %ax
   notw %ax
   orw %si, %ax
   movw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWByteNOTDST_OR_VAL
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteDST_OR_VAL:     { dst := dst or ValueW }
   orw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWByteDST_OR_VAL
   jmp .Lseg_memchangeMMXValueWQuad
.Lseg_memchangeMMXValueWByteSET:            { dst := 1 }
   movw $0xFFFF, %ax
   rep
   stosw

.Lseg_memchangeMMXValueWQuad:
   movl %esi, %eax            { restore eax }
   movl %edx, %ecx            { ecx = size div 4 }
   shrl $2, %ecx
   jz .Lseg_memchangeMMXValueWFinish
   jmpl .Lseg_memchangeMMXValueWQuadTable(, %ebx, 4)
.Lseg_memchangeMMXValueWQuadTable:
.long .Lseg_memchangeMMXValueWQuadCLEAR          { dst := 0 }
.long .Lseg_memchangeMMXValueWQuadNOT_DSTORVAL   { dst := not (dst or src) }
.long .Lseg_memchangeMMXValueWQuadDST_AND_NOTVAL { dst := dst and (not src) }
.long .Lseg_memchangeMMXValueWQuadNOTVAL         { dst := not src }
.long .Lseg_memchangeMMXValueWQuadVAL_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeMMXValueWQuadNOTDST         { dst := not dst }
.long .Lseg_memchangeMMXValueWQuadDST_XOR_VAL    { dst := dst xor src }
.long .Lseg_memchangeMMXValueWQuadNOT_DSTANDVAL  { dst := not (dst and src) }
.long .Lseg_memchangeMMXValueWQuadDST_AND_VAL    { dst := dst and src }
.long .Lseg_memchangeMMXValueWQuadNOT_DSTXORVAL  { dst := not (dst xor src) }
.long .Lseg_memchangeMMXValueWQuadDST            { dst := dst; do nothing }
.long .Lseg_memchangeMMXValueWQuadDST_OR_NOTVAL  { dst := dst or (not src) }
.long .Lseg_memchangeMMXValueWQuadVAL            { dst := src; copy }
.long .Lseg_memchangeMMXValueWQuadNOTDST_OR_VAL  { dst := (not dst) or src }
.long .Lseg_memchangeMMXValueWQuadDST_OR_VAL     { dst := dst or src }
.long .Lseg_memchangeMMXValueWQuadSET            { dst := 1 }

.Lseg_memchangeMMXValueWQuadCLEAR:          { dst := 0 }
   movq MMXQuadClearW, %mm0
.LValueWQuadCLEAR:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueWQuadCLEAR
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadNOT_DSTORVAL:   { dst := not (dst or ValueW) }
   movq MMXQuadVarW, %mm0
   por  %es:(%edi), %mm0
   pandn MMXQuadSetW, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWQuadNOT_DSTORVAL
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadDST_AND_NOTVAL: { dst := dst and (not ValueW) }
   movq MMXQuadVarW, %mm0
   pandn MMXQuadSetW, %mm0
   movq %mm0, MMXQuadVarW
.LValueWQuadDST_AND_NOTVAL:
   movq MMXQuadVarW, %mm0
   pand %es:(%edi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueWQuadDST_AND_NOTVAL
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadNOTVAL:         { dst := not ValueW }
   movq MMXQuadVarW, %mm0
   pandn MMXQuadSetW, %mm0
.LValueWQuadNOTVAL:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueWQuadNOTVAL
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadVAL_AND_NOTDST: { dst := ValueW and (not dst) }
   movq %es:(%edi), %mm0
   pandn MMXQuadSetW, %mm0
   pand MMXQuadVarW, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWQuadVAL_AND_NOTDST
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadNOTDST:         { dst := not dst }
   movq %es:(%edi), %mm0
   pandn MMXQuadSetW, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWQuadNOTDST
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadDST_XOR_VAL:    { dst := dst xor ValueW }
   movq %es:(%edi), %mm0
   pxor MMXQuadVarW, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWQuadDST_XOR_VAL
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadNOT_DSTANDVAL:  { dst := not (dst and ValueW) }
   movq %es:(%edi), %mm0
   pandl MMXQuadVarW, %mm0
   pandn MMXQuadSetW, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWQuadNOT_DSTANDVAL
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadDST_AND_VAL:    { dst := dst and ValueW }
   movq %es:(%edi), %mm0
   pand MMXQuadVarW, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWQuadDST_AND_VAL
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadNOT_DSTXORVAL:  { dst := not (dst xor ValueW) }
   movq %es:(%edi), %mm0
   pxor MMXQuadVarW, %mm0
   pandn MMXQuadSetW, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWQuadNOT_DSTXORVAL
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadDST_OR_NOTVAL:  { dst := dst or (not ValueW) }
   movq MMXQuadVarW, %mm0
   pandn MMXQuadSetW, %mm0
.LValueWQuadDST_OR_NOTVAL:
   movq %mm0, %mm1
   por %es:(%edi), %mm1
   movq %mm1, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueWQuadDST_OR_NOTVAL
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadVAL:            { dst := ValueW; fill }
   movq MMXQuadVarW, %mm0
.LValueWQuadVAL:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueWQuadVAL
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadNOTDST_OR_VAL:  { dst := (not dst) or ValueW }
   movq %es:(%edi), %mm0
   pandn MMXQuadSetW, %mm0
   por MMXQuadVarW, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWQuadNOTDST_OR_VAL
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadDST_OR_VAL:     { dst := dst or ValueW }
   movq %es:(%edi), %mm0
   por MMXQuadVarW, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueWQuadDST_OR_VAL
   jmp .Lseg_memchangeMMXValueWFinish
.Lseg_memchangeMMXValueWQuadSET:            { dst := 1 }
   movq MMXQuadSetW, %mm0
.LValueWQuadSET:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueWQuadSET

.Lseg_memchangeMMXValueWFinish:
   popw %es
   emms
.Lseg_memchangeMMXValueWExit:
   leave
   ret $16
MMXQuadSetW:
   .quad 0xffffffffffffffff { MMX reg with all bits set }
MMXQuadClearW:
   .quad 0x0000000000000000 { MMX reg with all bits clear }
MMXQuadVarW:
   .quad 0x1234567890abcdef { MMX reg to store value in }
end;


{ does a boolean operation between memory and a dword-sized value; MMX version }
procedure seg_memchangeValueDMMX (sel : Word; ofs : DWord; size : DWord; value : DWord; op : Mem_Op); assembler;
asm
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .Lseg_memchangeMMXValueDExit  { if (size = 0) then exit; }
   movw SEL, %bx
   movl OFS, %edi
   pushw %es
   movw %bx, %es           { dst = es:(edi) }
   { expand dword to MMX quad }
   movl VALUE, %eax
   movl %eax, MMXQuadVarD
   movl %eax, 4+MMXQuadVarD

   cld                     { clear direction flag for forward copy }

   movl %eax, %esi         { save value }
   movl OP, %ebx
   movl %ecx, %edx         { save size }
   andl $1, %ecx
   jz .Lseg_memchangeMMXValueDQuad { if (size and $1) <> 0 then }
   jmpl .Lseg_memchangeMMXValueDByteTable(, %ebx, 4) { case (op) of }
.Lseg_memchangeMMXValueDByteTable:
.long .Lseg_memchangeMMXValueDByteCLEAR          { dst := 0 }
.long .Lseg_memchangeMMXValueDByteNOT_DSTORVAL   { dst := not (dst or src) }
.long .Lseg_memchangeMMXValueDByteDST_AND_NOTVAL { dst := dst and (not src) }
.long .Lseg_memchangeMMXValueDByteNOTVAL         { dst := not src }
.long .Lseg_memchangeMMXValueDByteVAL_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeMMXValueDByteNOTDST         { dst := not dst }
.long .Lseg_memchangeMMXValueDByteDST_XOR_VAL    { dst := dst xor src }
.long .Lseg_memchangeMMXValueDByteNOT_DSTANDVAL  { dst := not (dst and src) }
.long .Lseg_memchangeMMXValueDByteDST_AND_VAL    { dst := dst and src }
.long .Lseg_memchangeMMXValueDByteNOT_DSTXORVAL  { dst := not (dst xor src) }
.long .Lseg_memchangeMMXValueDByteDST            { dst := dst; do nothing }
.long .Lseg_memchangeMMXValueDByteDST_OR_NOTVAL  { dst := dst or (not src) }
.long .Lseg_memchangeMMXValueDByteVAL            { dst := src; copy }
.long .Lseg_memchangeMMXValueDByteNOTDST_OR_VAL  { dst := (not dst) or src }
.long .Lseg_memchangeMMXValueDByteDST_OR_VAL     { dst := dst or src }
.long .Lseg_memchangeMMXValueDByteSET            { dst := 1 }

.Lseg_memchangeMMXValueDByteCLEAR:          { dst := 0 }
   xorl %eax, %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteNOT_DSTORVAL:   { dst := not (dst or ValueD) }
   movl %esi, %eax
   orl  %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDByteNOT_DSTORVAL
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteDST_AND_NOTVAL: { dst := dst and (not ValueD) }
   notl %eax
.LMMXValueDByteDST_AND_NOTVAL:
   andl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .LMMXValueDByteDST_AND_NOTVAL
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteNOTVAL:         { dst := not ValueD }
   notl %eax
   rep
   stosl
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteVAL_AND_NOTDST: { dst := ValueD and (not dst) }
   movl %es:(%edi), %eax
   notl %eax
   andl %esi, %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDByteVAL_AND_NOTDST
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteNOTDST:         { dst := not dst }
   notl %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDByteNOTDST
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteDST_XOR_VAL:    { dst := dst xor ValueD }
   xorl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDByteDST_XOR_VAL
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteNOT_DSTANDVAL:  { dst := not (dst and ValueD) }
   movl %esi, %eax
   andl %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDByteNOT_DSTANDVAL
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteDST_AND_VAL:    { dst := dst and ValueD }
   andl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDByteDST_AND_VAL
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteNOT_DSTXORVAL:  { dst := not (dst xor ValueD) }
   movl %esi, %eax
   xorl %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDByteNOT_DSTXORVAL
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDByteDST_OR_NOTVAL:  { dst := dst or (not ValueD) }
   notl %eax
.LMMXValueDByteDST_OR_NOTVAL:
   orl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .LMMXValueDByteDST_OR_NOTVAL
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteVAL:            { dst := ValueD; fill }
   rep
   stosl
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteNOTDST_OR_VAL:  { dst := (not dst) or ValueD }
   movl %es:(%edi), %eax
   notl %eax
   orl %esi, %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDByteNOTDST_OR_VAL
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteDST_OR_VAL:     { dst := dst or ValueD }
   orl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDByteDST_OR_VAL
   jmp .Lseg_memchangeMMXValueDQuad
.Lseg_memchangeMMXValueDByteSET:            { dst := 1 }
   movl $0xFFFFFFFF, %eax
   movl %eax, %es:(%edi)
   addl $4, %edi

.Lseg_memchangeMMXValueDQuad:
   movl %esi, %eax            { restore eax }
   movl %edx, %ecx            { ecx = size div 2 }
   shrl $1, %ecx
   jz .Lseg_memchangeMMXValueDFinish
   jmpl .Lseg_memchangeMMXValueDQuadTable(, %ebx, 4)
.Lseg_memchangeMMXValueDQuadTable:
.long .Lseg_memchangeMMXValueDQuadCLEAR          { dst := 0 }
.long .Lseg_memchangeMMXValueDQuadNOT_DSTORVAL   { dst := not (dst or src) }
.long .Lseg_memchangeMMXValueDQuadDST_AND_NOTVAL { dst := dst and (not src) }
.long .Lseg_memchangeMMXValueDQuadNOTVAL         { dst := not src }
.long .Lseg_memchangeMMXValueDQuadVAL_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeMMXValueDQuadNOTDST         { dst := not dst }
.long .Lseg_memchangeMMXValueDQuadDST_XOR_VAL    { dst := dst xor src }
.long .Lseg_memchangeMMXValueDQuadNOT_DSTANDVAL  { dst := not (dst and src) }
.long .Lseg_memchangeMMXValueDQuadDST_AND_VAL    { dst := dst and src }
.long .Lseg_memchangeMMXValueDQuadNOT_DSTXORVAL  { dst := not (dst xor src) }
.long .Lseg_memchangeMMXValueDQuadDST            { dst := dst; do nothing }
.long .Lseg_memchangeMMXValueDQuadDST_OR_NOTVAL  { dst := dst or (not src) }
.long .Lseg_memchangeMMXValueDQuadVAL            { dst := src; copy }
.long .Lseg_memchangeMMXValueDQuadNOTDST_OR_VAL  { dst := (not dst) or src }
.long .Lseg_memchangeMMXValueDQuadDST_OR_VAL     { dst := dst or src }
.long .Lseg_memchangeMMXValueDQuadSET            { dst := 1 }

.Lseg_memchangeMMXValueDQuadCLEAR:          { dst := 0 }
   movq MMXQuadClearD, %mm0
.LValueDQuadCLEAR:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueDQuadCLEAR
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadNOT_DSTORVAL:   { dst := not (dst or ValueD) }
   movq MMXQuadVarD, %mm0
   por  %es:(%edi), %mm0
   pandn MMXQuadSetD, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDQuadNOT_DSTORVAL
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadDST_AND_NOTVAL: { dst := dst and (not ValueD) }
   movq MMXQuadVarD, %mm0
   pandn MMXQuadSetD, %mm0
   movq %mm0, MMXQuadVarD
.LValueDQuadDST_AND_NOTVAL:
   movq MMXQuadVarD, %mm0
   pand %es:(%edi), %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueDQuadDST_AND_NOTVAL
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadNOTVAL:         { dst := not ValueD }
   movq MMXQuadVarD, %mm0
   pandn MMXQuadSetD, %mm0
.LValueDQuadNOTVAL:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueDQuadNOTVAL
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadVAL_AND_NOTDST: { dst := ValueD and (not dst) }
   movq %es:(%edi), %mm0
   pandn MMXQuadSetD, %mm0
   pand MMXQuadVarD, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDQuadVAL_AND_NOTDST
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadNOTDST:         { dst := not dst }
   movq %es:(%edi), %mm0
   pandn MMXQuadSetD, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDQuadNOTDST
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadDST_XOR_VAL:    { dst := dst xor ValueD }
   movq %es:(%edi), %mm0
   pxor MMXQuadVarD, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDQuadDST_XOR_VAL
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadNOT_DSTANDVAL:  { dst := not (dst and ValueD) }
   movq %es:(%edi), %mm0
   pandl MMXQuadVarD, %mm0
   pandn MMXQuadSetD, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDQuadNOT_DSTANDVAL
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadDST_AND_VAL:    { dst := dst and ValueD }
   movq %es:(%edi), %mm0
   pand MMXQuadVarD, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDQuadDST_AND_VAL
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadNOT_DSTXORVAL:  { dst := not (dst xor ValueD) }
   movq %es:(%edi), %mm0
   pxor MMXQuadVarD, %mm0
   pandn MMXQuadSetD, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDQuadNOT_DSTXORVAL
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadDST_OR_NOTVAL:  { dst := dst or (not ValueD) }
   movq MMXQuadVarD, %mm0
   pandn MMXQuadSetD, %mm0
.LValueDQuadDST_OR_NOTVAL:
   movq %mm0, %mm1
   por %es:(%edi), %mm1
   movq %mm1, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueDQuadDST_OR_NOTVAL
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadVAL:            { dst := ValueD; fill }
   movq MMXQuadVarD, %mm0
.LValueDQuadVAL:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueDQuadVAL
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadNOTDST_OR_VAL:  { dst := (not dst) or ValueD }
   movq %es:(%edi), %mm0
   pandn MMXQuadSetD, %mm0
   por MMXQuadVarD, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDQuadNOTDST_OR_VAL
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadDST_OR_VAL:     { dst := dst or ValueD }
   movq %es:(%edi), %mm0
   por MMXQuadVarD, %mm0
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .Lseg_memchangeMMXValueDQuadDST_OR_VAL
   jmp .Lseg_memchangeMMXValueDFinish
.Lseg_memchangeMMXValueDQuadSET:            { dst := 1 }
   movq MMXQuadSetD, %mm0
.LValueDQuadSET:
   movq %mm0, %es:(%edi)
   addl $8, %edi
   decl %ecx
   jnz .LValueDQuadSET

.Lseg_memchangeMMXValueDFinish:
   popw %es
   emms
.Lseg_memchangeMMXValueDExit:
   leave
   ret $18
MMXQuadSetD:
   .quad 0xffffffffffffffff { MMX reg with all bits set }
MMXQuadClearD:
   .quad 0x0000000000000000 { MMX reg with all bits clear }
MMXQuadVarD:
   .quad 0x1234567890abcdef { MMX reg to store value in }
end;

{ -------------------------------------------------------------------------- }
{ STD versions of :                                                          }
{                  seg_memchange()            (last rev. 15.10.1998)         }
{                  seg_memchangeValueB()      (last rev. 15.10.1998)         }
{                  seg_memchangeValueW()      (last rev. 15.10.1998)         }
{                  seg_memchangeValueD()      (last rev. 15.10.1998)         }
{ -------------------------------------------------------------------------- }

{ does a boolean operation between two memory areas; MMX version }
procedure seg_memchangeSTD (srcsel : Word; srcofs : DWord; dstsel : Word; dstofs : DWord; size : DWord; op : Mem_Op); assembler;
asm
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .Lseg_memchangeExit  { if (size = 0) then exit; }
   movw SRCSEL, %bx
   movw DSTSEL, %dx
   movl SRCOFS, %esi
   movl DSTOFS, %edi
   movl OP, %eax
   pushw %ds
   pushw %es
   movw %bx, %ds           { src = ds:(esi) }
   movw %dx, %es           { dst = es:(edi) }

   cld                     { clear direction flag for forward copy }

   movl %eax, %edx         { save size }
   andl $3, %ecx
   jz .Lseg_memchangeDWord { if (size and $3) <> 0 then }
   jmpl .Lseg_memchangeByteTable(, %ebx, 4) { case (op) of }
.Lseg_memchangeByteTable:
.long .Lseg_memchangeByteCLEAR          { dst := 0 }
.long .Lseg_memchangeByteNOT_DSTORSRC   { dst := not (dst or src) }
.long .Lseg_memchangeByteDST_AND_NOTSRC { dst := dst and (not src) }
.long .Lseg_memchangeByteNOTSRC         { dst := not src }
.long .Lseg_memchangeByteSRC_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeByteNOTDST         { dst := not dst }
.long .Lseg_memchangeByteDST_XOR_SRC    { dst := dst xor src }
.long .Lseg_memchangeByteNOT_DSTANDSRC  { dst := not (dst and src) }
.long .Lseg_memchangeByteDST_AND_SRC    { dst := dst and src }
.long .Lseg_memchangeByteNOT_DSTXORSRC  { dst := not (dst xor src) }
.long .Lseg_memchangeByteDST            { dst := dst; do nothing }
.long .Lseg_memchangeByteDST_OR_NOTSRC  { dst := dst or (not src) }
.long .Lseg_memchangeByteSRC            { dst := src; copy }
.long .Lseg_memchangeByteNOTDST_OR_SRC  { dst := (not dst) or src }
.long .Lseg_memchangeByteDST_OR_SRC     { dst := dst or src }
.long .Lseg_memchangeByteSET            { dst := 1 }

.Lseg_memchangeByteCLEAR:          { dst := 0 }
   xorb %al, %al
   rep
   stosb
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteNOT_DSTORSRC:   { dst := not (dst or src) }
   movb (%esi), %al
   orb  %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeByteNOT_DSTORSRC
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteDST_AND_NOTSRC: { dst := dst and (not src) }
   movb (%esi), %al
   notb %al
   andb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeByteDST_AND_NOTSRC
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteNOTSRC:         { dst := not src }
   movb (%esi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeByteNOTSRC
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteSRC_AND_NOTDST: { dst := src and (not dst) }
   movb %es:(%edi), %al
   notb %al
   andb (%esi), %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeByteSRC_AND_NOTDST
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteNOTDST:         { dst := not dst }
   notb %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeByteNOTDST
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteDST_XOR_SRC:    { dst := dst xor src }
   movb (%esi), %al
   xorb %al, %es:(%edi)
   incl %edi
   incl %esi
   decl %ecx
   jnz .Lseg_memchangeByteDST_XOR_SRC
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteNOT_DSTANDSRC:  { dst := not (dst and src) }
   movb (%esi), %al
   andb %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeByteNOT_DSTANDSRC
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteDST_AND_SRC:    { dst := dst and src }
   movb (%esi), %al
   andb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeByteDST_AND_SRC
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteNOT_DSTXORSRC:  { dst := not (dst xor src) }
   movb (%esi), %al
   xorb %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeByteNOT_DSTXORSRC
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeFinish
.Lseg_memchangeByteDST_OR_NOTSRC:  { dst := dst or (not src) }
   movb (%esi), %al
   notb %al
   orb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeByteDST_OR_NOTSRC
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteSRC:            { dst := src; copy }
   rep
   movsb
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteNOTDST_OR_SRC:  { dst := (not dst) or src }
   movb %es:(%edi), %al
   notb %al
   orb (%esi), %al
   movb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeByteNOTDST_OR_SRC
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteDST_OR_SRC:     { dst := dst or src }
   movb (%esi), %al
   orb %al, %es:(%edi)
   incl %esi
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeByteDST_OR_SRC
   jmp .Lseg_memchangeDWord
.Lseg_memchangeByteSET:            { dst := 1 }
   movb $0xFF, %al
   rep
   stosb

.Lseg_memchangeDWord:
   movl %edx, %ecx            { ecx = size div 4 }
   shrl $2, %ecx
   jz .Lseg_memchangeFinish
   jmpl .Lseg_memchangeDWordTable(, %ebx, 4)
.Lseg_memchangeDWordTable:
.long .Lseg_memchangeDWordCLEAR          { dst := 0 }
.long .Lseg_memchangeDWordNOT_DSTORSRC   { dst := not (dst or src) }
.long .Lseg_memchangeDWordDST_AND_NOTSRC { dst := dst and (not src) }
.long .Lseg_memchangeDWordNOTSRC         { dst := not src }
.long .Lseg_memchangeDWordSRC_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeDWordNOTDST         { dst := not dst }
.long .Lseg_memchangeDWordDST_XOR_SRC    { dst := dst xor src }
.long .Lseg_memchangeDWordNOT_DSTANDSRC  { dst := not (dst and src) }
.long .Lseg_memchangeDWordDST_AND_SRC    { dst := dst and src }
.long .Lseg_memchangeDWordNOT_DSTXORSRC  { dst := not (dst xor src) }
.long .Lseg_memchangeDWordDST            { dst := dst; do nothing }
.long .Lseg_memchangeDWordDST_OR_NOTSRC  { dst := dst or (not src) }
.long .Lseg_memchangeDWordSRC            { dst := src; copy }
.long .Lseg_memchangeDWordNOTDST_OR_SRC  { dst := (not dst) or src }
.long .Lseg_memchangeDWordDST_OR_SRC     { dst := dst or src }
.long .Lseg_memchangeDWordSET            { dst := 1 }

.Lseg_memchangeDWordCLEAR:          { dst := 0 }
   xorl %eax, %eax
   rep
   stosl
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordNOT_DSTORSRC:   { dst := not (dst or src) }
   movl (%esi), %eax
   orl  %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %esi
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordNOT_DSTORSRC
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordDST_AND_NOTSRC: { dst := dst and (not src) }
   movl (%esi), %eax
   notl %eax
   andl %eax, %es:(%edi)
   addl $4, %esi
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordDST_AND_NOTSRC
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordNOTSRC:         { dst := not src }
   movl (%esi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %esi
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordNOTSRC
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordSRC_AND_NOTDST: { dst := src and (not dst) }
   movl %es:(%edi), %eax
   notl %eax
   andl (%esi), %eax
   movl %eax, %es:(%edi)
   addl $4, %esi
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordSRC_AND_NOTDST
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordNOTDST:         { dst := not dst }
   notl %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordNOTDST
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordDST_XOR_SRC:    { dst := dst xor src }
   movl (%esi), %eax
   xorl %eax, %es:(%edi)
   addl $4, %esi
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordDST_XOR_SRC
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordNOT_DSTANDSRC:  { dst := not (dst and src) }
   movl (%esi), %eax
   andl %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %esi
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordNOT_DSTANDSRC
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordDST_AND_SRC:    { dst := dst and src }
   movl (%esi), %eax
   andl %eax, %es:(%edi)
   addl $4, %esi
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordDST_AND_SRC
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordNOT_DSTXORSRC:  { dst := not (dst xor src) }
   movl (%esi), %eax
   xorl %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %esi
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordNOT_DSTXORSRC
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordDST_OR_NOTSRC:  { dst := dst or (not src) }
   movl (%esi), %eax
   notl %eax
   orl %eax, %es:(%edi)
   addl $4, %esi
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordDST_OR_NOTSRC
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordSRC:            { dst := src; copy }
   rep
   movsl
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordNOTDST_OR_SRC:  { dst := (not dst) or src }
   movl %es:(%edi), %eax
   notl %eax
   orl (%esi), %eax
   movl %eax, %es:(%edi)
   addl $4, %esi
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordNOTDST_OR_SRC
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordDST_OR_SRC:     { dst := dst or src }
   movl (%esi), %eax
   orl %eax, %es:(%edi)
   addl $4, %esi
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeDWordDST_OR_SRC
   jmp .Lseg_memchangeFinish
.Lseg_memchangeDWordSET:            { dst := 1 }
   movl $0xFFFFFFFF, %eax
   rep
   stosl

.Lseg_memchangeFinish:
   popw %es
   popw %ds
.Lseg_memchangeExit:
end;

{ does a boolean operation between memory and a byte-sized value; STD version }
procedure seg_memchangeValueBSTD (sel : Word; ofs : DWord; size : DWord; value : Byte; op : Mem_Op); assembler;
asm
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .Lseg_memchangeValueExit  { if (size = 0) then exit; }
   movw SEL, %bx
   movl OFS, %edi
   pushw %es
   movw %bx, %es           { dst = es:(edi) }
   { expand char to dword }
   movb VALUE, %al
   movb %al, %ah
   movw %ax, %bx
   roll $16, %eax
   movw %bx, %ax

   cld                     { clear direction flag for forward copy }

   movl %eax, %esi         { save expanded value }
   movl OP, %ebx
   movl %ecx, %edx         { save size }
   andl $3, %ecx
   jz .Lseg_memchangeValueDWord { if (size and $3) <> 0 then }
   jmpl .Lseg_memchangeValueByteTable(, %ebx, 4) { case (op) of }
.Lseg_memchangeValueByteTable:
.long .Lseg_memchangeValueByteCLEAR          { dst := 0 }
.long .Lseg_memchangeValueByteNOT_DSTORVAL   { dst := not (dst or src) }
.long .Lseg_memchangeValueByteDST_AND_NOTVAL { dst := dst and (not src) }
.long .Lseg_memchangeValueByteNOTVAL         { dst := not src }
.long .Lseg_memchangeValueByteVAL_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeValueByteNOTDST         { dst := not dst }
.long .Lseg_memchangeValueByteDST_XOR_VAL    { dst := dst xor src }
.long .Lseg_memchangeValueByteNOT_DSTANDVAL  { dst := not (dst and src) }
.long .Lseg_memchangeValueByteDST_AND_VAL    { dst := dst and src }
.long .Lseg_memchangeValueByteNOT_DSTXORVAL  { dst := not (dst xor src) }
.long .Lseg_memchangeValueByteDST            { dst := dst; do nothing }
.long .Lseg_memchangeValueByteDST_OR_NOTVAL  { dst := dst or (not src) }
.long .Lseg_memchangeValueByteVAL            { dst := src; copy }
.long .Lseg_memchangeValueByteNOTDST_OR_VAL  { dst := (not dst) or src }
.long .Lseg_memchangeValueByteDST_OR_VAL     { dst := dst or src }
.long .Lseg_memchangeValueByteSET            { dst := 1 }

.Lseg_memchangeValueByteCLEAR:          { dst := 0 }
   xorb %al, %al
   rep
   stosb
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteNOT_DSTORVAL:   { dst := not (dst or value) }
   movb %ah, %al
   orb  %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeValueByteNOT_DSTORVAL
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteDST_AND_NOTVAL: { dst := dst and (not value) }
   notb %al
.LValueByteDST_AND_NOTVAL:
   andb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .LValueByteDST_AND_NOTVAL
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteNOTVAL:         { dst := not value }
   notb %al
   rep
   stosb
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteVAL_AND_NOTDST: { dst := value and (not dst) }
   movb %es:(%edi), %al
   notb %al
   andb %ah, %al
   movb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeValueByteVAL_AND_NOTDST
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteNOTDST:         { dst := not dst }
   notb %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeValueByteNOTDST
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteDST_XOR_VAL:    { dst := dst xor value }
   xorb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeValueByteDST_XOR_VAL
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteNOT_DSTANDVAL:  { dst := not (dst and value) }
   movb %ah, %al
   andb %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeValueByteNOT_DSTANDVAL
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteDST_AND_VAL:    { dst := dst and value }
   andb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeValueByteDST_AND_VAL
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteNOT_DSTXORVAL:  { dst := not (dst xor value) }
   movb %ah, %al
   xorb %es:(%edi), %al
   notb %al
   movb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeValueByteNOT_DSTXORVAL
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueByteDST_OR_NOTVAL:  { dst := dst or (not value) }
   notb %al
.LValueByteDST_OR_NOTVAL:
   orb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .LValueByteDST_OR_NOTVAL
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteVAL:            { dst := value; fill }
   rep
   stosb
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteNOTDST_OR_VAL:  { dst := (not dst) or value }
   movb %es:(%edi), %al
   notb %al
   orb %ah, %al
   movb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeValueByteNOTDST_OR_VAL
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteDST_OR_VAL:     { dst := dst or value }
   orb %al, %es:(%edi)
   incl %edi
   decl %ecx
   jnz .Lseg_memchangeValueByteDST_OR_VAL
   jmp .Lseg_memchangeValueDWord
.Lseg_memchangeValueByteSET:            { dst := 1 }
   movb $0xFF, %al
   rep
   stosb

.Lseg_memchangeValueDWord:
   movl %esi, %eax            { restore eax }
   movl %edx, %ecx            { ecx = size div 4 }
   shrl $2, %ecx
   jz .Lseg_memchangeValueFinish
   jmpl .Lseg_memchangeValueDWordTable(, %ebx, 4)
.Lseg_memchangeValueDWordTable:
.long .Lseg_memchangeValueDWordCLEAR          { dst := 0 }
.long .Lseg_memchangeValueDWordNOT_DSTORVAL   { dst := not (dst or src) }
.long .Lseg_memchangeValueDWordDST_AND_NOTVAL { dst := dst and (not src) }
.long .Lseg_memchangeValueDWordNOTVAL         { dst := not src }
.long .Lseg_memchangeValueDWordVAL_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeValueDWordNOTDST         { dst := not dst }
.long .Lseg_memchangeValueDWordDST_XOR_VAL    { dst := dst xor src }
.long .Lseg_memchangeValueDWordNOT_DSTANDVAL  { dst := not (dst and src) }
.long .Lseg_memchangeValueDWordDST_AND_VAL    { dst := dst and src }
.long .Lseg_memchangeValueDWordNOT_DSTXORVAL  { dst := not (dst xor src) }
.long .Lseg_memchangeValueDWordDST            { dst := dst; do nothing }
.long .Lseg_memchangeValueDWordDST_OR_NOTVAL  { dst := dst or (not src) }
.long .Lseg_memchangeValueDWordVAL            { dst := src; copy }
.long .Lseg_memchangeValueDWordNOTDST_OR_VAL  { dst := (not dst) or src }
.long .Lseg_memchangeValueDWordDST_OR_VAL     { dst := dst or src }
.long .Lseg_memchangeValueDWordSET            { dst := 1 }

.Lseg_memchangeValueDWordCLEAR:          { dst := 0 }
   xorl %eax, %eax
   rep
   stosl
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordNOT_DSTORVAL:   { dst := not (dst or value) }
   movl %esi, %eax
   orl  %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDWordNOT_DSTORVAL
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordDST_AND_NOTVAL: { dst := dst and (not value) }
   notl %eax
.LValueDWordDST_AND_NOTVAL:
   andl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .LValueDWordDST_AND_NOTVAL
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordNOTVAL:         { dst := not value }
   notl %eax
   rep
   stosl
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordVAL_AND_NOTDST: { dst := value and (not dst) }
   movl %es:(%edi), %eax
   notl %eax
   andl %esi, %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDWordVAL_AND_NOTDST
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordNOTDST:         { dst := not dst }
   notl %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDWordNOTDST
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordDST_XOR_VAL:    { dst := dst xor value }
   xorl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDWordDST_XOR_VAL
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordNOT_DSTANDVAL:  { dst := not (dst and value) }
   movl %esi, %eax
   andl %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDWordNOT_DSTANDVAL
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordDST_AND_VAL:    { dst := dst and value }
   andl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDWordDST_AND_VAL
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordNOT_DSTXORVAL:  { dst := not (dst xor value) }
   movl %esi, %eax
   xorl %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDWordNOT_DSTXORVAL
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordDST_OR_NOTVAL:  { dst := dst or (not value) }
   notl %eax
.LValueDWordDST_OR_NOTVAL:
   orl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .LValueDWordDST_OR_NOTVAL
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordVAL:            { dst := value; fill }
   rep
   stosl
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordNOTDST_OR_VAL:  { dst := (not dst) or value }
   movl %es:(%edi), %eax
   notl %eax
   orl  %esi, %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDWordNOTDST_OR_VAL
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordDST_OR_VAL:     { dst := dst or value }
   orl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDWordDST_OR_VAL
   jmp .Lseg_memchangeValueFinish
.Lseg_memchangeValueDWordSET:            { dst := 1 }
   movl $0xFFFFFFFF, %eax
   rep
   stosl

.Lseg_memchangeValueFinish:
   popw %es
.Lseg_memchangeValueExit:
end;

{ does a boolean operation between memory and a word-sized value; STD version }
procedure seg_memchangeValueWSTD (sel : Word; ofs : DWord; size : DWord; value : Word; op : Mem_Op); assembler;
asm
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .Lseg_memchangeValueWExit  { if (size = 0) then exit; }
   movw SEL, %bx
   movl OFS, %edi
   pushw %es
   movw %bx, %es           { dst = es:(edi) }
   { expand word to dword }
   movw VALUE, %ax
   roll $16, %eax
   movw VALUE, %ax

   cld                     { clear direction flag for forward copy }

   movl %eax, %esi         { save expanded value }
   movl OP, %ebx
   movl %ecx, %edx         { save size }
   andl $1, %ecx
   jz .Lseg_memchangeValueWDWord { if (size and $1) <> 0 then }
   jmpl .Lseg_memchangeValueWByteTable(, %ebx, 4) { case (op) of }
.Lseg_memchangeValueWByteTable:
.long .Lseg_memchangeValueWByteCLEAR          { dst := 0 }
.long .Lseg_memchangeValueWByteNOT_DSTORVAL   { dst := not (dst or src) }
.long .Lseg_memchangeValueWByteDST_AND_NOTVAL { dst := dst and (not src) }
.long .Lseg_memchangeValueWByteNOTVAL         { dst := not src }
.long .Lseg_memchangeValueWByteVAL_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeValueWByteNOTDST         { dst := not dst }
.long .Lseg_memchangeValueWByteDST_XOR_VAL    { dst := dst xor src }
.long .Lseg_memchangeValueWByteNOT_DSTANDVAL  { dst := not (dst and src) }
.long .Lseg_memchangeValueWByteDST_AND_VAL    { dst := dst and src }
.long .Lseg_memchangeValueWByteNOT_DSTXORVAL  { dst := not (dst xor src) }
.long .Lseg_memchangeValueWByteDST            { dst := dst; do nothing }
.long .Lseg_memchangeValueWByteDST_OR_NOTVAL  { dst := dst or (not src) }
.long .Lseg_memchangeValueWByteVAL            { dst := src; copy }
.long .Lseg_memchangeValueWByteNOTDST_OR_VAL  { dst := (not dst) or src }
.long .Lseg_memchangeValueWByteDST_OR_VAL     { dst := dst or src }
.long .Lseg_memchangeValueWByteSET            { dst := 1 }

.Lseg_memchangeValueWByteCLEAR:          { dst := 0 }
   xorw %ax, %ax
   rep
   stosw
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteNOT_DSTORVAL:   { dst := not (dst or value) }
   movw %si, %ax
   orw  %es:(%edi), %ax
   notw %ax
   movw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWByteNOT_DSTORVAL
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteDST_AND_NOTVAL: { dst := dst and (not value) }
   notw %ax
.LValueWByteDST_AND_NOTVAL:
   andw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .LValueWByteDST_AND_NOTVAL
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteNOTVAL:         { dst := not value }
   notw %ax
   rep
   stosw
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteVAL_AND_NOTDST: { dst := value and (not dst) }
   movw %es:(%edi), %ax
   notw %ax
   andw %si, %ax
   movw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWByteVAL_AND_NOTDST
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteNOTDST:         { dst := not dst }
   notw %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWByteNOTDST
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteDST_XOR_VAL:    { dst := dst xor value }
   xorw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWByteDST_XOR_VAL
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteNOT_DSTANDVAL:  { dst := not (dst and value) }
   movw %si, %ax
   andw %es:(%edi), %ax
   notw %ax
   movw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWByteNOT_DSTANDVAL
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteDST_AND_VAL:    { dst := dst and value }
   andw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWByteDST_AND_VAL
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteNOT_DSTXORVAL:  { dst := not (dst xor value) }
   movw %si, %ax
   xorw %es:(%edi), %ax
   notw %ax
   movw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWByteNOT_DSTXORVAL
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWByteDST_OR_NOTVAL:  { dst := dst or (not value) }
   notw %ax
.LValueWByteDST_OR_NOTVAL:
   orw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .LValueWByteDST_OR_NOTVAL
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteVAL:            { dst := value; fill }
   rep
   stosw
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteNOTDST_OR_VAL:  { dst := (not dst) or value }
   movw %es:(%edi), %ax
   notw %ax
   orw %si, %ax
   movw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWByteNOTDST_OR_VAL
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteDST_OR_VAL:     { dst := dst or value }
   orw %ax, %es:(%edi)
   addl $2, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWByteDST_OR_VAL
   jmp .Lseg_memchangeValueWDWord
.Lseg_memchangeValueWByteSET:            { dst := 1 }
   movw $0xFFFF, %ax
   rep
   stosw

.Lseg_memchangeValueWDWord:
   movl %esi, %eax            { restore eax }
   movl %edx, %ecx            { ecx = size div 4 }
   shrl $1, %ecx
   jz .Lseg_memchangeValueWFinish
   jmpl .Lseg_memchangeValueWDWordTable(, %ebx, 4)
.Lseg_memchangeValueWDWordTable:
.long .Lseg_memchangeValueWDWordCLEAR          { dst := 0 }
.long .Lseg_memchangeValueWDWordNOT_DSTORVAL   { dst := not (dst or src) }
.long .Lseg_memchangeValueWDWordDST_AND_NOTVAL { dst := dst and (not src) }
.long .Lseg_memchangeValueWDWordNOTVAL         { dst := not src }
.long .Lseg_memchangeValueWDWordVAL_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeValueWDWordNOTDST         { dst := not dst }
.long .Lseg_memchangeValueWDWordDST_XOR_VAL    { dst := dst xor src }
.long .Lseg_memchangeValueWDWordNOT_DSTANDVAL  { dst := not (dst and src) }
.long .Lseg_memchangeValueWDWordDST_AND_VAL    { dst := dst and src }
.long .Lseg_memchangeValueWDWordNOT_DSTXORVAL  { dst := not (dst xor src) }
.long .Lseg_memchangeValueWDWordDST            { dst := dst; do nothing }
.long .Lseg_memchangeValueWDWordDST_OR_NOTVAL  { dst := dst or (not src) }
.long .Lseg_memchangeValueWDWordVAL            { dst := src; copy }
.long .Lseg_memchangeValueWDWordNOTDST_OR_VAL  { dst := (not dst) or src }
.long .Lseg_memchangeValueWDWordDST_OR_VAL     { dst := dst or src }
.long .Lseg_memchangeValueWDWordSET            { dst := 1 }

.Lseg_memchangeValueWDWordCLEAR:          { dst := 0 }
   xorl %eax, %eax
   rep
   stosl
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordNOT_DSTORVAL:   { dst := not (dst or value) }
   movl %esi, %eax
   orl  %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWDWordNOT_DSTORVAL
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordDST_AND_NOTVAL: { dst := dst and (not value) }
   notl %eax
.LValueWDWordDST_AND_NOTVAL:
   andl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .LValueWDWordDST_AND_NOTVAL
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordNOTVAL:         { dst := not value }
   notl %eax
   rep
   stosl
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordVAL_AND_NOTDST: { dst := value and (not dst) }
   movl %es:(%edi), %eax
   notl %eax
   andl %esi, %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWDWordVAL_AND_NOTDST
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordNOTDST:         { dst := not dst }
   notl %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWDWordNOTDST
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordDST_XOR_VAL:    { dst := dst xor value }
   xorl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWDWordDST_XOR_VAL
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordNOT_DSTANDVAL:  { dst := not (dst and value) }
   movl %esi, %eax
   andl %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWDWordNOT_DSTANDVAL
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordDST_AND_VAL:    { dst := dst and value }
   andl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWDWordDST_AND_VAL
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordNOT_DSTXORVAL:  { dst := not (dst xor value) }
   movl %esi, %eax
   xorl %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWDWordNOT_DSTXORVAL
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordDST_OR_NOTVAL:  { dst := dst or (not value) }
   notl %eax
.LValueWDWordDST_OR_NOTVAL:
   orl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .LValueWDWordDST_OR_NOTVAL
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordVAL:            { dst := value; fill }
   rep
   stosl
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordNOTDST_OR_VAL:  { dst := (not dst) or value }
   movl %es:(%edi), %eax
   notl %eax
   orl  %esi, %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWDWordNOTDST_OR_VAL
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordDST_OR_VAL:     { dst := dst or value }
   orl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueWDWordDST_OR_VAL
   jmp .Lseg_memchangeValueWFinish
.Lseg_memchangeValueWDWordSET:            { dst := 1 }
   movl $0xFFFFFFFF, %eax
   rep
   stosl

.Lseg_memchangeValueWFinish:
   popw %es
.Lseg_memchangeValueWExit:
end;

{ does a boolean operation between memory and a dword-sized value; STD version }
procedure seg_memchangeValueDSTD (sel : Word; ofs : DWord; size : DWord; value : DWord; op : Mem_Op); assembler;
asm
   movl SIZE, %ecx
   testl %ecx, %ecx
   jz .Lseg_memchangeValueDExit  { if (size = 0) then exit; }
   movw SEL, %bx
   movl OFS, %edi
   pushw %es
   movw %bx, %es           { dst = es:(edi) }
   cld                     { clear direction flag for forward copy }
   movl VALUE, %eax
   movl %eax, %esi         { save value }

   movl OP, %ebx
   jmpl .Lseg_memchangeValueDDWordTable(, %ebx, 4)
.Lseg_memchangeValueDDWordTable:
.long .Lseg_memchangeValueDDWordCLEAR          { dst := 0 }
.long .Lseg_memchangeValueDDWordNOT_DSTORVAL   { dst := not (dst or src) }
.long .Lseg_memchangeValueDDWordDST_AND_NOTVAL { dst := dst and (not src) }
.long .Lseg_memchangeValueDDWordNOTVAL         { dst := not src }
.long .Lseg_memchangeValueDDWordVAL_AND_NOTDST { dst := src and (not dst) }
.long .Lseg_memchangeValueDDWordNOTDST         { dst := not dst }
.long .Lseg_memchangeValueDDWordDST_XOR_VAL    { dst := dst xor src }
.long .Lseg_memchangeValueDDWordNOT_DSTANDVAL  { dst := not (dst and src) }
.long .Lseg_memchangeValueDDWordDST_AND_VAL    { dst := dst and src }
.long .Lseg_memchangeValueDDWordNOT_DSTXORVAL  { dst := not (dst xor src) }
.long .Lseg_memchangeValueDDWordDST            { dst := dst; do nothing }
.long .Lseg_memchangeValueDDWordDST_OR_NOTVAL  { dst := dst or (not src) }
.long .Lseg_memchangeValueDDWordVAL            { dst := src; copy }
.long .Lseg_memchangeValueDDWordNOTDST_OR_VAL  { dst := (not dst) or src }
.long .Lseg_memchangeValueDDWordDST_OR_VAL     { dst := dst or src }
.long .Lseg_memchangeValueDDWordSET            { dst := 1 }

.Lseg_memchangeValueDDWordCLEAR:          { dst := 0 }
   xorl %eax, %eax
   rep
   stosl
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordNOT_DSTORVAL:   { dst := not (dst or value) }
   movl %esi, %eax
   orl  %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDDWordNOT_DSTORVAL
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordDST_AND_NOTVAL: { dst := dst and (not value) }
   notl %eax
.LValueDDWordDST_AND_NOTVAL:
   andl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .LValueDDWordDST_AND_NOTVAL
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordNOTVAL:         { dst := not value }
   notl %eax
   rep
   stosl
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordVAL_AND_NOTDST: { dst := value and (not dst) }
   movl %es:(%edi), %eax
   notl %eax
   andl %esi, %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDDWordVAL_AND_NOTDST
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordNOTDST:         { dst := not dst }
   notl %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDDWordNOTDST
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordDST_XOR_VAL:    { dst := dst xor value }
   xorl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDDWordDST_XOR_VAL
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordNOT_DSTANDVAL:  { dst := not (dst and value) }
   movl %esi, %eax
   andl %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDDWordNOT_DSTANDVAL
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordDST_AND_VAL:    { dst := dst and value }
   andl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDDWordDST_AND_VAL
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordNOT_DSTXORVAL:  { dst := not (dst xor value) }
   movl %esi, %eax
   xorl %es:(%edi), %eax
   notl %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDDWordNOT_DSTXORVAL
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordDST:            { dst := dst; do nothing }
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordDST_OR_NOTVAL:  { dst := dst or (not value) }
   notl %eax
.LValueDDWordDST_OR_NOTVAL:
   orl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .LValueDDWordDST_OR_NOTVAL
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordVAL:            { dst := value; fill }
   rep
   stosl
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordNOTDST_OR_VAL:  { dst := (not dst) or value }
   movl %es:(%edi), %eax
   notl %eax
   orl  %esi, %eax
   movl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDDWordNOTDST_OR_VAL
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordDST_OR_VAL:     { dst := dst or value }
   orl %eax, %es:(%edi)
   addl $4, %edi
   decl %ecx
   jnz .Lseg_memchangeValueDDWordDST_OR_VAL
   jmp .Lseg_memchangeValueDFinish
.Lseg_memchangeValueDDWordSET:            { dst := 1 }
   movl $0xFFFFFFFF, %eax
   rep
   stosl

.Lseg_memchangeValueDFinish:
   popw %es
.Lseg_memchangeValueDExit:
end;

{ Other routines }

procedure memcpy (var src, dst; size : DWord);
begin
     seg_memcpy (mydseg, longint(@src), mydseg, longint(@dst), size);
end;

procedure memchange (var src, dst; size : DWord; op : Mem_Op);
begin
     seg_memchange (mydseg, longint(@src), mydseg, longint(@dst), size, op);
end;

procedure memchangeValueB (var x; size : DWord; value : Byte; op : Mem_Op);
begin
     seg_memchangeValueB (mydseg, longint(@x), size, value, op);
end;

procedure memchangeValueW (var x; size : DWord; value : Word; op : Mem_Op);
begin
     seg_memchangeValueW (mydseg, longint(@x), size, value, op);
end;

procedure memchangeValueD (var x; size : DWord; value : DWord; op : Mem_Op);
begin
     seg_memchangeValueD (mydseg, longint(@x), size, value, op);
end;

procedure memsetb (var x; size : DWord; value : Byte);
begin
     seg_memsetb (mydseg, longint(@x), size, value);
end;

procedure memsetW (var x; size : DWord; value : Word);
begin
     seg_memsetW (mydseg, longint(@x), size, value);
end;

procedure memsetD (var x; size : DWord; value : DWord);
begin
   seg_memsetD (mydseg, longint(@x), size, value);
end;

begin
     {$ASMMODE ATT}
     asm
        movw %ds, %ax
        movw %ax, mydseg
     end ['EAX'];

     { set all to standard operations first }
     seg_memcpy  := @seg_memcpy_std;
     seg_memsetb := @seg_memsetb_std;
     seg_memsetw := @seg_memsetw_std;
     seg_memsetd := @seg_memsetd_std;
     seg_memchange := @seg_memchangestd;
     seg_memchangevalueb := @seg_memchangevaluebstd;
     seg_memchangevaluew := @seg_memchangevaluewstd;
     seg_memchangevalued := @seg_memchangevaluedstd;

     { overwrite standard routines with the specialized ones if available }
     { first check for better memcpy() routines }

     {$IFDEF USE_KNI}
     if (cpu_supported(cHAVE_KNI)) then begin
     	{$IFDEF DEBUG}
        Writeln(stderr, 'MEMORY : Using experimental KNI for memcpy()');
        {$ENDIF}
        seg_memcpy := @seg_memcpy_kni;
     end else {$ENDIF USE_KNI} if (cpu_supported(cHAVE_MMX)) then begin
     	{$IFDEF DEBUG}
        Writeln(stderr, 'MEMORY : Using MMX for memcpy()');
        {$ENDIF}
        seg_memcpy := @seg_memcpy_mmx;
     end else if (this_cpu.manufacturer = cMFC_Intel) then begin
     	{$IFDEF DEBUG}
        Writeln(stderr, 'MEMORY : Using FPU for memcpy()');
        {$ENDIF}
        seg_memcpy := @seg_memcpy_fpu;
     end;
     { then check if other procedures can be substituted too }
     if (cpu_supported(cHAVE_MMX)) then begin
        {$IFDEF DEBUG}
        Writeln(stderr, 'MEMORY : Using MMX for memsetX()');
        Writeln(stderr, 'MEMORY : Using MMX for memchange()');
        Writeln(stderr, 'MEMORY : Using MMX for memchangevalueX()');
        {$ENDIF}
        seg_memsetb := @seg_memsetb_mmx;
        seg_memsetw := @seg_memsetw_mmx;
        seg_memsetd := @seg_memsetd_mmx;
        seg_memchange := @seg_memchangeMMX;
        seg_memchangevalueb := @seg_memchangevaluebMMX;
        seg_memchangevaluew := @seg_memchangevaluewMMX;
        seg_memchangevalued := @seg_memchangevaluedMMX;
     end;
     {$IFDEF DEBUG}
     Writeln(stderr);
     {$ENDIF}
end.
{ 02-01-1999 start of history, release 2 finished (revision 1)
  - complete rewrite of all functions except the ROP's
  - now the memcpy automatically aligns on source
  - all memsetX do address align & use largest available memory set
  instructions to fill block, giving a speed increase of up to 4 times RTL
  functions
  - adjusted MEMORY to use CPU...
  - changed name of xxx_memset() functions to xxx_memsetb() to have more
  uniform names
  - automatically writes comments to stderr if compiler directive DEBUG is
  enabled

  06-01-1999 small fixes
  - now allows smartlinking (from MarcoV@stack.nl)
  - removed FPU memory copy code, because certain byte combinations may(?)
  cause a FPU exception (reported by muller@cerbere.u-strasbg.fr). Actually I
  never encountered such problem, so I only commented the FPU detection code.

  15-02-1999 updated FPU code
  - now using fildq and fistpq instructions to move memory on non-MMX intel
  processors (to avoid possible (but never reported) FPU exceptions). This
  was the original code, but somehow the fp-equivalents were used by me.

  23-03-1999 some KNI experimental code
  - added seg_memcpy() for KNI instruction set; still unsure about sense of
  this (I can't test if it's worth), i.e. if there's any speed improvement
  over MMX instructions......

  30-03-1999 code check
  - removed unnecessary KNI warning and wait for return key at startup
  - detection code of available CPU features rewritten (with the old KNI code
  was never used since all KNI supporting CPUs support MMX too.

  13-05-1999 adaption to new assembler
  - added a $OUTPUT_FORMAT AS to the implementation section
  - works again
}
